class DefaultMasterInputDict:
    """
    DefaultMasterInput is a class that handles all the default values
    for the master input dictionary.
    """

    def __init__(self):
        """
        This constructor creates the defaults master input dictionary.
        This is called default_input
        """
        self.default_input_dict = dict()

        # Set the list and dictionary values on the master dictionary
        self.default_input_dict['season_construct'] = ['spring', 'summer', 'fall']
        self.default_input_dict['time_construct'] = 'normal'
        self.default_input_dict['hour_day'] = {'long': 24, 'normal': 10}
        self.default_input_dict['operational_construction_time'] = self.default_input_dict['hour_day'][
            self.default_input_dict['time_construct']]

    def populate_input_dict(self, incomplete_input_dict):
        """
        Completely fills the input_dict. If there are any keys in the
        default input dictionary that are not on incomplete_input_dict, those
        missing key/value pairs are placed on a fully populated input
        dictionary.

        This method uses the self.__default_input_dict for the values
        it fills in the output dictionary.

        Parameters
        ----------
        incomplete_input_dict : dict
            The dict of input values that may or may not be complete.

        Returns
        -------
        dict
            The fully populated output diction
        """
        complete_dict = dict()
        for key, value in incomplete_input_dict.items():
            complete_dict[key] = value
        for key, value in self.default_input_dict.items():
            if key not in complete_dict:
                complete_dict[key] = value
        return complete_dict



import os
import sys

# Numpy deprecation warnings
import warnings

import numpy as np

from wisdem.inputs import load_yaml
from wisdem.inputs.gui import run as guirun
from wisdem.glue_code.runWISDEM import run_wisdem

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
warnings.simplefilter("ignore", RuntimeWarning, lineno=175)
warnings.simplefilter("ignore", RuntimeWarning, lineno=177)

def read_master_file(fyaml):
    if os.path.exists(fyaml):
        print("...Reading master input file,", fyaml)
    else:
        raise FileNotFoundError("The master input file, " + fyaml + ", cannot be found.")

    input_yaml = load_yaml(fyaml)

    check_list = ["geometry_file", "modeling_file", "analysis_file"]
    for f in check_list:
        if not os.path.exists(input_yaml[f]):
            raise FileNotFoundError("The " + f + " entry, " + input_yaml[f] + ", cannot be found.")

    return input_yaml


def wisdem_cmd():
    usg_msg = "WISDEM command line launcher\n    Arguments: \n    wisdem : Starts GUI\n    wisdem input.yaml : Runs master yaml file that specifies geometry, modeling, and analysis files\n    wisdem geom.yaml modeling.yaml analysis.yaml : Runs specific geometry, modeling, and analysis files\n"

    # Look for help message
    help_flag = False
    for k in range(len(sys.argv)):
        if sys.argv[k] in ["-h", "--help"]:
            help_flag = True

    if help_flag:
        print(usg_msg)

    elif len(sys.argv) == 1:
        # Launch GUI
        guirun()

    elif len(sys.argv) == 2:
        # Grab master input file
        fyaml = sys.argv[1]
        if os.path.exists(fyaml):
            print("...Reading master input file,", fyaml)
        else:
            raise FileNotFoundError("The master input file, " + fyaml + ", cannot be found.")
        yaml_dict = load_yaml(fyaml)

        check_list = ["geometry_file", "modeling_file", "analysis_file"]
        for f in check_list:
            if not os.path.exists(yaml_dict[f]):
                raise FileNotFoundError("The " + f + " entry, " + yaml_dict[f] + ", cannot be found.")

        # Run WISDEM (also saves output)
        wt_opt, modeling_options, opt_options = run_wisdem(
            yaml_dict["geometry_file"], yaml_dict["modeling_file"], yaml_dict["analysis_file"]
        )

    elif len(sys.argv) == 4:
        check_list = ["geometry", "modeling", "analysis"]
        for k, f in enumerate(sys.argv[1:]):
            if not os.path.exists(f):
                raise FileNotFoundError("The " + check_list[k] + " file, " + f + ", cannot be found.")

        # Run WISDEM (also saves output)
        wt_opt, modeling_options, opt_options = run_wisdem(sys.argv[1], sys.argv[2], sys.argv[3])

    else:
        # As if asked for help
        print("Unrecognized set of inputs.  Usage:")
        print(usg_msg)

    sys.exit(0)


if __name__ == "__main__":
    wisdem_cmd()



"""Provides the `ExportSystemDesign` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"

import numpy as np

from wisdem.orbit.phases.design._cables import CableSystem


class ExportSystemDesign(CableSystem):
    """
    Design phase for the export cabling system.

    Attributes
    ----------
    num_cables : int
        Total number of cables required for transmitting power.
    length : float
        Length of a single cable connecting the OSS to the interconnection in km.
    mass : float
        Mass of `length` in tonnes.
    cable : `Cable`
        Instance of `ORBIT.phases.design.Cable`. An export system will
        only require a single type of cable.
    total_length : float
        Total length of cable required to trasmit power.
    total_mass : float
        Total mass of cable required to transmit power.
    sections_cables : np.ndarray, shape: (`num_cables, )
        An array of `cable`.
    sections_lengths : np.ndarray, shape: (`num_cables, )
        An array of `length`.
    """

    expected_config = {
        "site": {"distance_to_landfall": "km", "depth": "m"},
        "landfall": {"interconnection_distance": "km (optional)"},
        "plant": {"capacity": "MW"},
        "export_system_design": {
            "cables": "str",
            "num_redundant": "int (optional)",
            "touchdown_distance": "m (optional, default: 0)",
            "percent_added_length": "float (optional)",
        },
    }

    output_config = {
        "export_system": {
            "cable": {
                "linear_density": "t/km",
                "number": "int",
                "sections": "list",
                "cable_power": "MW",
            }
        }
    }

    def __init__(self, config, **kwargs):
        """
        Defines the cables and sections required to install an offshore wind
        farm.

        Parameters
        ----------
        config : dict
            Configuration dictionary for the array cabling system. See
            `expected_config` for details on what is required.
        """

        super().__init__(config, "export", **kwargs)
        # For export cables there should only be one cable type due to the
        # custom nature of their design

        for name in self.expected_config["site"]:
            setattr(self, "".join(("_", name)), config["site"][name])
        self._depth = config["site"]["depth"]
        self._plant_capacity = self.config["plant"]["capacity"]
        self._distance_to_landfall = config["site"]["distance_to_landfall"]
        self._get_touchdown_distance()
        try:
            self._distance_to_interconnection = config["landfall"]["interconnection_distance"]
        except KeyError:
            self._distance_to_interconnection = 3

    def run(self):
        """
        Instantiates the export cable system and runs all the required methods.
        """

        self._initialize_cables()
        self.cable = self.cables[[*self.cables][0]]
        self.compute_number_cables()
        self.compute_cable_length()
        self.compute_cable_mass()
        self.compute_total_cable()

    @property
    def total_cable_cost(self):
        """Returns total array system cable cost."""

        return sum(self.cost_by_type.values())

    @property
    def detailed_output(self):
        """Returns export system design outputs."""

        _output = {
            **self.design_result,
            "export_system_total_mass": self.total_mass,
            "export_system_total_length": self.total_length,
            "export_system_total_cost": self.total_cable_cost,
            "export_system_cable_power": self.cable.cable_power,
        }

        return _output

    def compute_number_cables(self):
        """
        Calculate the total number of required and redundant cables to
        transmit power to the onshore interconnection.
        """

        num_required = np.ceil(self._plant_capacity / self.cable.cable_power)
        num_redundant = self._design.get("num_redundant", 0)

        self.num_cables = int(num_required + num_redundant)

    def compute_cable_length(self):
        """
        Calculates the total distance an export cable must travel.
        """

        added_length = 1.0 + self._design.get("percent_added_length", 0.0)
        self.length = round(
            (
                self.free_cable_length
                + (self._distance_to_landfall - self.touchdown / 1000)
                + self._distance_to_interconnection
            )
            * added_length,
            10,
        )

    def compute_cable_mass(self):
        """
        Calculates the total mass of a single length of export cable.
        """

        self.mass = round(self.length * self.cable.linear_density, 10)

    def compute_total_cable(self):
        """
        Calculates the total length and mass of cables required to fully
        connect the OSS to the interconnection point.
        """

        self.total_length = round(self.num_cables * self.length, 10)
        self.total_mass = round(self.num_cables * self.mass, 10)

    @property
    def sections_cable_lengths(self):
        """
        Creates an array of section lengths to work with `CableSystem`

        Returns
        -------
        np.ndarray
            Array of `length` with shape (`num_cables`, ).
        """
        return np.full(self.num_cables, self.length)

    @property
    def sections_cables(self):
        """
        Creates an array of cable names to work with `CableSystem`.

        Returns
        -------
        np.ndarray
            Array of `cable.name` with shape (`num_cables`, ).
        """

        return np.full(self.num_cables, self.cable.name)

    @property
    def design_result(self):
        """
        A dictionary of cables types and number of different cable lengths and
        linear density.

        Returns
        -------
        output : dict
            Dictionary containing the output export system. Contains:
            - 'linear_density': 't/km'
            - 'sections': 'list [self.length]'
            - 'number': 'int'
        """

        if self.cables is None:
            raise Exception(f"Has {self.__class__.__name__} been ran?")

        output = {
            "export_system": {
                "interconnection_distance": self._distance_to_interconnection,
                "system_cost": self.total_cost,
            }
        }

        for name, cable in self.cables.items():
            output["export_system"]["cable"] = {
                "linear_density": cable.linear_density,
                "sections": [self.length],
                "number": self.num_cables,
                "cable_power": cable.cable_power,
            }

        return output



__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2022, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = ["jake.nunemaker@nrel.gov"]


from copy import deepcopy

from benedict import benedict

from wisdem.orbit import ProjectManager

DEFAULT_MULTIPLIERS = {
    "blades": {"domestic": 0.026, "imported": 0.30},
    "nacelle": {"domestic": 0.025, "imported": 0.10},
    "tower": {
        "domestic": 0.04,
        "imported": 0.20,
        "tariffs": 0.25,
    },
    "monopile": {
        "domestic": 0.085,
        "imported": 0.28,
        "tariffs": 0.25,
    },
    "transition_piece": {
        "domestic": 0.169,
        "imported": 0.17,
        "tariffs": 0.25,
    },
    "array_cable": {"domestic": 0.19, "imported": 0.0},
    "export_cable": {"domestic": 0.231, "imported": 0.0},
    "oss_topside": {"domestic": 0.0, "imported": 0.0},
    "oss_substructure": {"domestic": 0.0, "imported": 0.0},
}


TURBINE_CAPEX_SPLIT = {"blades": 0.135, "nacelle": 0.274, "tower": 0.162}


LABOR_SPLIT = {"tower": 0.5, "monopile": 0.5, "transition_piece": 0.5, "oss_topside": 0.5}


class SupplyChainManager:
    def __init__(self, supply_chain_configuration, **kwargs):
        """
        Creates an instance of `SupplyChainManager`.

        Parameters
        ----------
        configuration : dict
        """

        self.sc_config = supply_chain_configuration
        self.multipliers = kwargs.get("multipliers", DEFAULT_MULTIPLIERS)
        self.labor_split = kwargs.get("labor_split", LABOR_SPLIT)
        self.turbine_split = kwargs.get("turbine_split", TURBINE_CAPEX_SPLIT)

    def run_project(self, config, weather=None, **kwargs):
        """
        Runs an ORBIT configuration, performing any supply chain related pre
        and post processing steps on either side of the ORBIT run.

        Parameters
        ----------
        config : dict
            ORBIT Configuration
        weather : pd.DataFrame | None
            Weather data to use for the ORBIT simulation.
        """

        config = benedict(deepcopy(config))
        config = self.pre_process(config)

        project = ProjectManager(config, weather=weather, **kwargs)
        project.run()

        project = self.post_process(project)

        return project

    def pre_process(self, config):
        """"""

        # Save original plant design
        plant = deepcopy(config["plant"])

        # Run ProjectManager without install phases to generate design results
        install_phases = config["install_phases"]
        config["install_phases"] = []
        project = ProjectManager(config)
        project.run()
        config = deepcopy(project.config)

        # Replace calculated plant design with original
        config["plant"] = plant

        # Run pre ORBIT supply chain adjustments
        config = self.process_turbine_capex(config)
        config = self.process_monopile_capex(config)
        config = self.process_transition_piece_capex(config)
        config = self.process_offshore_substation_topside_capex(config)

        # Add install phases back in
        config["install_phases"] = install_phases
        config["design_phases"] = []

        return config

    def post_process(self, project):
        """"""

        project = self.process_array_cable_capex(project)
        project = self.process_export_cable_capex(project)

        return project

    # Pre Processing Methods
    def process_turbine_capex(self, config):
        """
        Add blade, nacelle and tower cost adders to input `turbine_capex`.

        Parameters
        ----------
        config : dict
            ORBIT configuration.
        """

        blade_scenario = self.sc_config["blades"]
        nacelle_scenario = self.sc_config["nacelle"]
        tower_scenario = self.sc_config["blades"]

        blade_mult = self.multipliers["blades"].get(blade_scenario, None)
        if blade_mult == None:
            print(f"Warning: scenario '{blade_scenario}' not found for category 'blades'.")
            blade_mult = 0.0

        nacelle_mult = self.multipliers["nacelle"].get(nacelle_scenario, None)
        if nacelle_mult == None:
            print(f"Warning: scenario '{nacelle_scenario}' not found for category 'nacelle'.")
            nacelle_mult = 0.0

        raw_cost = config.get("project_parameters.turbine_capex", 1300)
        blade_adder = raw_cost * self.turbine_split["blades"] * blade_mult
        nacelle_adder = raw_cost * self.turbine_split["nacelle"] * nacelle_mult

        if tower_scenario == "domestic, imported steel":
            tower_adder = self.multipliers["tower"]["domestic"] * raw_cost
            tower_tariffs = (
                raw_cost
                * self.turbine_split["tower"]
                * (1 - self.labor_split["tower"])
                * self.multipliers["tower"]["tariffs"]
            )

        else:
            tower_tariffs = 0.0
            tower_mult = self.multipliers["tower"].get(tower_scenario, None)
            if tower_mult == None:
                print(f"Warning: scenario '{tower_scenario}' not found for category 'tower'.")
                tower_mult = 0.0

            tower_adder = raw_cost * self.turbine_split["tower"] * tower_mult

        config["project_parameters.turbine_capex"] = sum(
            [raw_cost, blade_adder, nacelle_adder, tower_adder, tower_tariffs]
        )

        return config

    def process_monopile_capex(self, config):
        """
        Add monopile cost adder and potential tariffs to `monopile.unit_cost`.

        Parameters
        ----------
        config : dict
            ORBIT configuration.
        """

        raw_cost = config["monopile.unit_cost"]
        scenario = self.sc_config["monopile"]

        if scenario == "domestic, imported steel":
            adder = self.multipliers["monopile"]["domestic"] * raw_cost
            tariffs = raw_cost * (1 - self.labor_split["monopile"]) * self.multipliers["monopile"]["tariffs"]

        else:
            tariffs = 0.0
            mult = self.multipliers["monopile"].get(scenario, None)
            if mult == None:
                print(f"Warning: scenario '{scenario}' not found for category 'monopile'.")
                mult = 0.0

            adder = raw_cost * mult

        config["monopile.unit_cost"] = sum([raw_cost, adder, tariffs])

        return config

    def process_transition_piece_capex(self, config):
        """
        Add transition piece cost adder and potential tariffs to
        `transition_piece.unit_cost`.

        Parameters
        ----------
        config : dict
            ORBIT configuration.
        """

        raw_cost = config["transition_piece.unit_cost"]
        scenario = self.sc_config["transition_piece"]

        if scenario == "domestic, imported steel":
            adder = self.multipliers["transition_piece"]["domestic"] * raw_cost
            tariffs = (
                raw_cost * (1 - self.labor_split["transition_piece"]) * self.multipliers["transition_piece"]["tariffs"]
            )

        else:
            tariffs = 0.0
            mult = self.multipliers["transition_piece"].get(scenario, None)
            if mult == None:
                print(f"Warning: scenario '{scenario}' not found for category 'transition_piece'.")
                mult = 0.0

            adder = raw_cost * mult

        config["transition_piece.unit_cost"] = sum([raw_cost, adder, tariffs])

        return config

    def process_offshore_substation_topside_capex(self, config):
        """
        Add OSS topside cost adder and potential tariffs to
        `offshore_substation_topside.unit_cost`.

        Parameters
        ----------
        config : dict
            ORBIT configuration.
        """

        raw_cost = config["offshore_substation_topside.unit_cost"]
        scenario = self.sc_config["oss_topside"]

        if scenario == "domestic, imported steel":
            adder = self.multipliers["oss_topside"]["domestic"] * raw_cost
            tariffs = raw_cost * (1 - self.labor_split["oss_topside"]) * self.multipliers["oss_topside"]["tariffs"]

        else:
            tariffs = 0.0
            mult = self.multipliers["oss_topside"].get(scenario, None)
            if mult == None:
                print(f"Warning: scenario '{scenario}' not found for category 'oss_topside'.")
                mult = 0.0

            adder = raw_cost * mult

        config["offshore_substation_topside.unit_cost"] = sum([raw_cost, adder, tariffs])

        return config

    # Post Processing Methods
    def process_array_cable_capex(self, project):
        """
        Add array cable cost adder.

        Parameters
        ----------
        project : ProjectManager
        """

        scenario = self.sc_config["array_cable"]
        mult = self.multipliers["array_cable"].get(scenario, None)
        if mult == None:
            print(f"Warning: scenario '{scenario}' not found for category 'array_cable'.")
            mult = 0.0

        project.system_costs["ArrayCableInstallation"] *= 1 + mult

        return project

    def process_export_cable_capex(self, project):
        """
        Add export cable cost adder.

        Parameters
        ----------
        project : ProjectManager
        """

        scenario = self.sc_config["export_cable"]
        mult = self.multipliers["export_cable"].get(scenario, None)
        if mult == None:
            print(f"Warning: scenario '{scenario}' not found for category 'export_cable'.")
            mult = 0.0

        project.system_costs["ExportCableInstallation"] *= 1 + mult

        return project



"""Tests for the `ArraySystemDesign` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


from copy import deepcopy

import numpy as np
import pytest

from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.phases.design import ArraySystemDesign, CustomArraySystemDesign
from wisdem.orbit.core.exceptions import LibraryItemNotFoundError

config_full_ring = extract_library_specs("config", "array_design_full_ring")

config_partial_ring = deepcopy(config_full_ring)
config_partial_ring["plant"]["num_turbines"] = 49

config_full_grid = deepcopy(config_full_ring)
config_full_grid["plant"]["layout"] = "grid"

config_partial_grid = deepcopy(config_full_grid)
config_partial_grid["plant"]["num_turbines"] = 49

config_cables_from_file_fail = deepcopy(config_full_grid)
config_cables_from_file_fail["array_system_design"]["cables"] = "Cable1"

config_custom_base = deepcopy(config_full_grid)
config_custom_base["plant"]["num_turbines"] = 8
config_custom_base["plant"]["layout"] = "custom"

config_incorrect_turbines = deepcopy(config_custom_base)
config_incorrect_turbines["plant"]["num_turbines"] = 11
config_incorrect_turbines["array_system_design"]["location_data"] = "passes"

config_missing_col = deepcopy(config_custom_base)
config_missing_col["array_system_design"]["location_data"] = "missing_columns"

config_incomplete_optional = deepcopy(config_custom_base)
config_incomplete_optional["array_system_design"]["location_data"] = "incomplete_optional"

config_incomplete_required = deepcopy(config_custom_base)
config_incomplete_required["array_system_design"]["location_data"] = "incomplete_required"

config_duplicate_coordinates = deepcopy(config_custom_base)
config_duplicate_coordinates["array_system_design"]["location_data"] = "duplicate_coordinates"


def test_array_system_creation():
    array = ArraySystemDesign(config_full_grid)
    array.run()
    assert array


def test_cable_not_found():
    array = ArraySystemDesign(config_cables_from_file_fail)
    with pytest.raises(LibraryItemNotFoundError):
        array.run()


@pytest.mark.parametrize(
    "config,num_full_strings,num_partial_strings,num_turbines_full_string,num_turbines_partial_string",
    (
        (config_full_ring, 10, 0, 4, 0),
        (config_partial_ring, 12, 1, 4, 1),
        (config_full_grid, 10, 0, 4, 0),
        (config_partial_grid, 12, 1, 4, 1),
    ),
    ids=["full_ring", "partial_ring", "full_grid", "partial_grid"],
)
def test_string_creation(
    config,
    num_full_strings,
    num_partial_strings,
    num_turbines_full_string,
    num_turbines_partial_string,
):
    array = ArraySystemDesign(config)
    array.run()

    assert array

    string_properties = (
        (num_full_strings, "num_full_strings"),
        (num_partial_strings, "num_partial_strings"),
        (num_turbines_full_string, "num_turbines_full_string"),
        (num_turbines_partial_string, "num_turbines_partial_string"),
    )
    for val, name in string_properties:
        assert val == array.__getattribute__(name)


@pytest.mark.parametrize(
    "turbine_rating,expected",
    ((1, 40), (2, 20), (4.5, 8), (6, 6), (8.5, 4), (12, 3)),
)
def test_max_turbines_per_cable(turbine_rating, expected):
    config = deepcopy(config_full_grid)
    config["array_system_design"]["cables"] = "XLPE_630mm_33kV"
    config["turbine"]["turbine_rating"] = turbine_rating
    array = ArraySystemDesign(config)
    array.run()
    assert array.cables["XLPE_630mm_33kV"].max_turbines == expected


@pytest.mark.parametrize(
    "config,shape,num_null",
    (
        (config_full_ring, (10, 4), 0),
        (config_partial_ring, (13, 4), 3),
        (config_full_grid, (10, 4), 0),
        (config_partial_grid, (13, 4), 3),
    ),
    ids=["full_ring", "partial_ring", "full_grid", "partial_grid"],
)
def test_grid_creation(config, shape, num_null):
    array = ArraySystemDesign(config)
    array.run()

    assert array.turbines_x.shape == shape
    assert array.turbines_y.shape == shape
    assert np.isnan(array.turbines_x).sum() == num_null
    assert np.isnan(array.turbines_y).sum() == num_null

    assert isinstance(array.oss_x, float)
    assert isinstance(array.oss_y, float)

    shape = (shape[0], shape[1] + 1, 2)
    assert array.coordinates.shape == shape
    assert np.isnan(array.coordinates).sum() == 2 * num_null


@pytest.mark.parametrize(
    "config,total_length",
    (
        (config_full_ring, 34.7),
        (config_partial_ring, 42.68),
        (config_full_grid, 53.29),
        (config_partial_grid, 77.0),
    ),
    ids=["full_ring", "partial_ring", "full_grid", "partial_grid"],
)
def test_total_cable_length(config, total_length):
    array = ArraySystemDesign(config)
    array.run()

    val = round(sum(val.sum() for val in array.cable_lengths_by_type.values()), 2)
    assert total_length == val

    val = round(sum(array.total_cable_length_by_type.values()), 2)
    assert total_length == val

    val = round(
        sum(
            length * n
            for val in array.design_result["array_system"]["cables"].values()
            for length, n in val["cable_sections"]
        ),
        2,
    )
    assert total_length == val


def test_missing_columns():
    array = CustomArraySystemDesign(config_missing_col)

    with pytest.raises(ValueError):
        array.run()


def test_duplicate_turbine_coordinates():
    array = CustomArraySystemDesign(config_duplicate_coordinates)

    with pytest.raises(ValueError):
        array.run()


def test_incomplete_required_columns():
    array = CustomArraySystemDesign(config_incomplete_required)

    with pytest.raises(ValueError):
        array.run()


def test_incomplete_optional_columns():
    array = CustomArraySystemDesign(config_incomplete_optional)

    with pytest.warns(UserWarning):
        array.run()


def test_correct_turbines():
    array = CustomArraySystemDesign(config_incorrect_turbines)

    with pytest.raises(ValueError):
        array.run()


def test_floating_calculations():
    base = deepcopy(config_full_ring)
    base["site"]["depth"] = 50
    number = base["plant"]["num_turbines"]

    sim = ArraySystemDesign(base)
    sim.run()

    base_length = sim.total_length

    floating_no_cat = deepcopy(base)
    floating_no_cat["site"]["depth"] = 250
    floating_no_cat["array_system_design"]["touchdown_distance"] = 0

    sim2 = ArraySystemDesign(floating_no_cat)
    sim2.run()

    no_cat_length = sim2.total_length
    assert no_cat_length == pytest.approx(base_length + 2 * (200 / 1000) * number)

    floating_cat = deepcopy(base)
    floating_cat["site"]["depth"] = 250

    sim3 = ArraySystemDesign(floating_cat)
    sim3.run()

    with_cat_length = sim3.total_length
    assert with_cat_length < no_cat_length



"""
Testing framework for the `ExportCableInstallation` class.
"""

__author__ = ["Rob Hammond", "Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


from copy import deepcopy

import pandas as pd
import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.phases.install import ExportCableInstallation
from wisdem.test.test_orbit.data import test_weather

base_config = extract_library_specs("config", "export_cable_install")
simul_config = deepcopy(base_config)
_ = simul_config.pop("export_cable_bury_vessel")


@pytest.mark.parametrize("config", (base_config, simul_config), ids=["separate", "simultaneous"])
def test_simulation_setup(config):
    sim = ExportCableInstallation(config)
    assert sim.env
    assert sim.cable
    assert sim.cable.linear_density

    actions = [a["action"] for a in sim.env.actions]
    assert "Onshore Construction" in actions


@pytest.mark.parametrize("config", (base_config, simul_config), ids=["separate", "simultaneous"])
def test_vessel_initialization(config):
    sim = ExportCableInstallation(config)
    assert sim.install_vessel
    assert sim.install_vessel.cable_storage

    if config.get("export_cable_bury_vessel", None):
        assert sim.bury_vessel


@pytest.mark.parametrize("config", (base_config, simul_config), ids=["separate", "simultaneous"])
@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
def test_for_complete_logging(config, weather):
    sim = ExportCableInstallation(config, weather=weather)
    sim.run()

    df = pd.DataFrame(sim.env.actions)
    df = df.loc[df["action"] != "Mobilize"].reset_index(drop=True)
    df = df.assign(shift=(df["time"] - df["time"].shift(1)))

    for vessel in df["agent"].unique():
        _df = df[df["agent"] == vessel].copy()
        _df = _df.assign(shift=(_df["time"] - _df["time"].shift(1)))
        assert (_df["shift"] - _df["duration"]).fillna(0.0).abs().max() < 1e-9

    assert ~df["cost"].isnull().any()
    _ = sim.agent_efficiencies
    _ = sim.detailed_output


def test_simultaneous_speed_kwargs():
    sim = ExportCableInstallation(simul_config)
    sim.run()
    baseline = sim.total_phase_time

    key = "cable_lay_bury_speed"
    val = pt[key] * 0.1

    kwargs = {key: val}

    sim = ExportCableInstallation(simul_config, **kwargs)
    sim.run()

    assert sim.total_phase_time > baseline


def test_separate_speed_kwargs():
    sim = ExportCableInstallation(base_config)
    sim.run()
    df = pd.DataFrame(sim.env.actions)

    base_lay = sum(df.loc[df["action"] == "Lay Cable"]["duration"])
    base_bury = sum(df.loc[df["action"] == "Bury Cable"]["duration"])

    kwargs = {
        "cable_lay_speed": pt["cable_lay_speed"] * 0.1,
        "cable_bury_speed": pt["cable_bury_speed"] * 0.1,
    }

    new = ExportCableInstallation(base_config, **kwargs)
    new.run()
    df = pd.DataFrame(new.env.actions)

    new_lay = sum(df.loc[df["action"] == "Lay Cable"]["duration"])
    assert new_lay > base_lay

    new_bury = sum(df.loc[df["action"] == "Bury Cable"]["duration"])
    assert new_bury > base_bury


def test_kwargs_for_export_install():
    new_export_system = {"cable": {"linear_density": 50.0, "sections": [1000], "number": 1}}
    new_site = {"distance": 50, "depth": 20}

    new_config = deepcopy(base_config)
    new_config["export_system"] = new_export_system
    new_config["site"] = new_site

    sim = ExportCableInstallation(new_config)
    sim.run()
    baseline = sim.total_phase_time

    keywords = [
        "onshore_construction_time",
        "cable_load_time",
        "site_position_time",
        "cable_prep_time",
        "cable_lower_time",
        "cable_pull_in_time",
        "cable_termination_time",
        "cable_splice_time",
        "tow_plow_speed",
        "pull_winch_speed",
        "cable_raise_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if "speed" in kw:
            _new = default - 0.05

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            kwargs = {kw: _new}

        else:
            kwargs = {kw: default + 2}

        new_sim = ExportCableInstallation(new_config, **kwargs)
        new_sim.run()
        new_time = new_sim.total_phase_time

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"ExpInstall: '{failed}' not affecting results.")

    else:
        assert True


def test_kwargs_for_export_install_in_ProjectManager():
    new_export_system = {
        "cable": {"linear_density": 50.0, "sections": [1000], "number": 1},
        "system_cost": 200e6,
    }
    new_site = {"distance": 50, "depth": 20}
    base = deepcopy(base_config)
    base["export_system"] = new_export_system
    base["site"] = new_site
    base["install_phases"] = ["ExportCableInstallation"]

    project = ProjectManager(base)
    project.run()
    baseline = project.phase_times["ExportCableInstallation"]

    keywords = [
        "onshore_construction_time",
        "cable_load_time",
        "site_position_time",
        "cable_prep_time",
        "cable_lower_time",
        "cable_pull_in_time",
        "cable_termination_time",
        "cable_splice_time",
        "tow_plow_speed",
        "pull_winch_speed",
        "cable_raise_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if "speed" in kw:
            _new = default - 0.05

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            processes = {kw: _new}

        else:
            processes = {kw: default + 2}

        new_config = deepcopy(base)
        new_config["processes"] = processes

        new_project = ProjectManager(new_config)
        new_project.run()
        new_time = new_project.phase_times["ExportCableInstallation"]

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"ExpInstall: '{failed}' not affecting results.")

    else:
        assert True



"""Core functionality of ORBIT installation phases."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from .port import Port, WetStorage
from .cargo import Cargo
from .vessel import Vessel
from .library import loader
from .components import Crane, JackingSys
from .environment import OrbitEnvironment as Environment
from .supply_chain import SubstructureDelivery



"""
Testing framework for common cable installation tasks.
"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


import pytest

from wisdem.orbit.core.exceptions import MissingComponent
from wisdem.orbit.phases.install.cable_install.common import (
    tow_plow,
    lay_cable,
    bury_cable,
    prep_cable,
    pull_winch,
    lower_cable,
    raise_cable,
    splice_cable,
    pull_in_cable,
    lay_bury_cable,
    terminate_cable,
    load_cable_on_vessel,
)


def test_load_cable_on_vessel(env, cable_vessel, feeder, simple_cable):
    env.register(cable_vessel)
    cable_vessel.initialize(mobilize=False)

    env.register(feeder)
    feeder.initialize(mobilize=False)

    load_cable_on_vessel(cable_vessel, simple_cable)
    env.run()

    with pytest.raises(MissingComponent):
        load_cable_on_vessel(feeder, simple_cable)
        env.run()


@pytest.mark.parametrize(
    "task, log, args",
    [
        (prep_cable, "Prepare Cable", []),
        (lower_cable, "Lower Cable", []),
        (pull_in_cable, "Pull In Cable", []),
        (terminate_cable, "Terminate Cable", []),
        (lay_bury_cable, "Lay/Bury Cable", [10]),
        (lay_cable, "Lay Cable", [10]),
        (splice_cable, "Splice Cable", []),
        (raise_cable, "Raise Cable", []),
        (tow_plow, "Tow Plow", [100]),
        (pull_winch, "Pull Winch", [100]),
    ],
)
def test_task(env, cable_vessel, task, log, args):
    env.register(cable_vessel)
    cable_vessel.initialize(mobilize=False)

    task(cable_vessel, *args)
    env.run()

    actions = [a["action"] for a in env.actions]
    assert log in actions



import copy
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.fixed_bottomse.monopile as mon
from wisdem.commonse import gravity as g

npts = 100
myones = np.ones((npts,))
RIGID = mon.RIGID


class TestPreDiscretization(unittest.TestCase):
    def testAll(self):
        inputs = {}
        outputs = {}
        inputs["monopile_height"] = 60.0
        inputs["monopile_foundation_height"] = -50.0
        inputs["tower_foundation_height"] = 10.0
        inputs["tower_base_diameter"] = 10.0
        inputs["monopile_top_diameter"] = 10.0
        inputs["water_depth"] = 30.0

        mydis = mon.PreDiscretization()
        mydis.compute(inputs, outputs)
        self.assertEqual(outputs["transition_piece_height"], 10.0)
        self.assertEqual(outputs["z_start"], -50.0)
        npt.assert_array_equal(outputs["joint1"], np.array([0.0, 0.0, -50.0]))
        npt.assert_array_equal(outputs["joint2"], np.array([0.0, 0.0, 10.0]))
        self.assertEqual(outputs["suctionpile_depth"], 20.0)
        self.assertEqual(outputs["s_const1"], 20.0 / 60.0)
        self.assertEqual(outputs["bending_height"], 40.0)
        self.assertEqual(outputs["constr_diam_consistency"], 1.0)

    def testBadHeight(self):
        inputs = {}
        outputs = {}
        inputs["monopile_height"] = 70.0
        inputs["monopile_foundation_height"] = -50.0
        inputs["tower_foundation_height"] = 10.0
        inputs["tower_base_diameter"] = 5.0
        inputs["monopile_top_diameter"] = 10.0
        inputs["water_depth"] = 30.0

        mydis = mon.PreDiscretization()
        mydis.compute(inputs, outputs)
        self.assertEqual(outputs["transition_piece_height"], 10.0)
        self.assertEqual(outputs["z_start"], -60.0)
        npt.assert_array_equal(outputs["joint1"], np.array([0.0, 0.0, -60.0]))
        npt.assert_array_equal(outputs["joint2"], np.array([0.0, 0.0, 10.0]))
        self.assertEqual(outputs["suctionpile_depth"], 30.0)
        self.assertEqual(outputs["s_const1"], 30.0 / 70.0)
        self.assertEqual(outputs["bending_height"], 40.0)
        self.assertEqual(outputs["constr_diam_consistency"], 0.5)


class TestMass(unittest.TestCase):
    def testAll(self):
        inputs = {}
        outputs = {}
        inputs["cylinder_mass"] = 10.0
        inputs["cylinder_cost"] = 100.0
        inputs["cylinder_z_cg"] = 50.0
        inputs["cylinder_I_base"] = 20 * np.ones(6)
        inputs["transition_piece_mass"] = 2.0
        inputs["transition_piece_cost"] = 5.0
        inputs["gravity_foundation_mass"] = 3.0
        inputs["tower_mass"] = 30.0
        inputs["tower_cost"] = 300.0
        inputs["z_full"] = np.linspace(-30, 10, 5)
        inputs["outer_diameter_full"] = 10.0 * np.ones(5)

        mydis = mon.MonopileMass(npts=5)
        mydis.compute(inputs, outputs)
        self.assertEqual(outputs["monopile_mass"], 10.0 + 2 + 3)
        self.assertEqual(outputs["monopile_cost"], 100 + 5)
        self.assertEqual(outputs["monopile_z_cg"], (10 * 50 + 2 * 10 - 3 * 30) / 15.0)
        self.assertEqual(outputs["structural_mass"], 10.0 + 2 + 3 + 30)
        self.assertEqual(outputs["structural_cost"], 100.0 + 5 + 300)
        I_root = np.r_[0.5 * 25, 0.5 * 25, 25, np.zeros(3)]
        npt.assert_equal(outputs["transition_piece_I"], 2 * I_root)
        npt.assert_equal(outputs["gravity_foundation_I"], 0.5 * 3 * I_root)
        npt.assert_equal(
            outputs["monopile_I_base"],
            20 * np.ones(6) + 0.5 * 3 * I_root + 2 * I_root + 2 * 40**2 * np.r_[1, 1, np.zeros(4)],
        )


class TestMonopileSE(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # Store analysis options
        self.modeling_options = {}
        self.modeling_options["materials"] = {}
        self.modeling_options["materials"]["n_mat"] = 1

        self.modeling_options["WISDEM"] = {}
        self.modeling_options["WISDEM"]["n_dlc"] = 1
        self.modeling_options["WISDEM"]["FixedBottomSE"] = {}
        self.modeling_options["WISDEM"]["FixedBottomSE"]["buckling_method"] = "eurocode"
        self.modeling_options["WISDEM"]["FixedBottomSE"]["buckling_length"] = 30.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = 0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = 0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_refine"] = 3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["wind"] = "PowerWind"

        self.modeling_options["WISDEM"]["FixedBottomSE"]["soil_springs"] = False
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gravity_foundation"] = False

        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_f"] = 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_m"] = 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_b"] = 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_fatigue"] = 1.0

        # Simplified the options available to the user
        self.modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"] = {}
        self.modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["shear"] = True
        self.modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["geom"] = True
        self.modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["modal_method"] = 1
        self.modeling_options["WISDEM"]["FixedBottomSE"]["frame3dd"]["tol"] = 1e-9
        self.modeling_options["WISDEM"]["FixedBottomSE"]["rank_and_file"] = False

        self.modeling_options["flags"] = {}
        self.modeling_options["flags"]["tower"] = False
        self.modeling_options["WISDEM"]["TowerSE"] = {}
        self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = 0

    def testProblemFixedPile(self):
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["FixedBottomSE"]["soil_springs"] = True
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gravity_foundation"] = False

        prob = om.Problem(reports=False)
        prob.model = mon.MonopileSE(modeling_options=self.modeling_options)
        prob.setup()

        prob["water_depth"] = 30.0
        prob["transition_piece_mass"] = 1e2
        prob["transition_piece_cost"] = 1e3
        prob["gravity_foundation_mass"] = 0.0  # 1e4

        hval = np.array([15.0, 30.0])
        prob["monopile_s"] = np.cumsum(np.r_[0, hval]) / hval.sum()
        prob["monopile_foundation_height"] = -45.0
        prob["monopile_height"] = hval.sum()
        prob["monopile_outer_diameter_in"] = 10.0 * np.ones(3)
        prob["monopile_top_diameter"] = 10.0
        prob["monopile_layer_thickness"] = 0.1 * np.ones((1, 3))
        prob["outfitting_factor_in"] = 1.0
        prob["monopile_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8
        prob["sigma_ult_mat"] = 1e8 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 1e1
        prob["wohler_A_mat"] = 1e1

        prob["yaw"] = 0.0
        prob["G_soil"] = 1e7
        prob["nu_soil"] = 0.5
        prob["wind_reference_height"] = 80.0
        prob["z0"] = 0.0
        prob["ca_usr"] = 0.0
        prob["cd_usr"] = -1.0
        prob["rho_air"] = 1.225
        prob["mu_air"] = 1.7934e-5
        prob["shearExp"] = 0.2
        prob["rho_water"] = 1025.0
        prob["mu_water"] = 1.3351e-3
        prob["beta_wind"] = prob["beta_wave"] = 0.0
        prob["Hsig_wave"] = 0.0
        prob["Tsig_wave"] = 1e3
        prob["env.Uref"] = 15.0
        prob["turbine_F"] = 1e3 * np.arange(2, 5)
        prob["turbine_M"] = 1e4 * np.arange(2, 5)
        prob["turbine_mass"] = 1e5
        prob["turbine_cg"] = 1e1 * np.ones(3)
        prob["turbine_I"] = 2e1 * np.ones(6)
        prob.run_model()

        # All other tests from above
        mass_dens = 1e4 * (5.0**2 - 4.9**2) * np.pi
        npt.assert_equal(prob["z_start"], -45.0)
        npt.assert_equal(prob["transition_piece_height"], 0.0)
        npt.assert_equal(prob["suctionpile_depth"], 15.0)
        npt.assert_equal(prob["z_param"], np.array([-45.0, -30.0, 0.0]))
        npt.assert_almost_equal(prob["monopile_mass"], mass_dens * 45.0 + 1e2)

        npt.assert_equal(prob.model.perf.monopile.frame.rnode, 1 + np.arange(4, dtype=np.int_))
        npt.assert_array_less(prob.model.perf.monopile.frame.rKx, RIGID)
        npt.assert_array_less(prob.model.perf.monopile.frame.rKy, RIGID)
        npt.assert_array_less(prob.model.perf.monopile.frame.rKz[0], RIGID)
        npt.assert_array_less(prob.model.perf.monopile.frame.rKtx, RIGID)
        npt.assert_array_less(prob.model.perf.monopile.frame.rKty, RIGID)
        npt.assert_array_less(prob.model.perf.monopile.frame.rKtz, RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKx, RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKy, RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKz[0], RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKtx, RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKty, RIGID)
        npt.assert_array_less(0.0, prob.model.perf.monopile.frame.rKtz, RIGID)
        npt.assert_equal(prob.model.perf.monopile.frame.rKz[1:], 0.0)

        npt.assert_equal(prob.model.perf.monopile.frame.ENMnode, np.array([6, 5, 1], dtype=np.int_))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMmass, np.array([1e5, 1e2, 0]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhox, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhoy, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhoz, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxx, np.array([20, 1e2 * 25 * 0.5, 0.0]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIyy, np.array([20, 1e2 * 25 * 0.5, 0.0]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIzz, np.array([20, 1e2 * 25, 0.0]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxy, np.r_[20.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxz, np.r_[20.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIyz, np.r_[20.0, 0, 0])
        self.assertEqual(prob.model.perf.monopile.frame.addGravityLoadForExtraNodeMass, [False, True, True])

        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].NF, np.array([6]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fx, np.array([2e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fy, np.array([3e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fz, np.array([4e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Mxx, np.array([2e4]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Myy, np.array([3e4]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Mzz, np.array([4e4]))

        # npt.assert_almost_equal(prob["monopile.mudline_F"], [4.61183362e04, 1.59353875e03, -2.94077236e07], 0)
        # npt.assert_almost_equal(prob["monopile.mudline_M"], [-248566.38259147, -3286049.81237828, 40000.0], 0)

    def testProblemFixedPile_GBF(self):
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["FixedBottomSE"]["soil_springs"] = False
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gravity_foundation"] = True

        prob = om.Problem(reports=False)
        prob.model = mon.MonopileSE(modeling_options=self.modeling_options)
        prob.setup()

        prob["water_depth"] = 30.0
        prob["transition_piece_mass"] = 1e2
        prob["transition_piece_cost"] = 1e3
        prob["gravity_foundation_mass"] = 1e4

        hval = np.array([15.0, 30.0])
        prob["monopile_s"] = np.cumsum(np.r_[0, hval]) / hval.sum()
        prob["monopile_foundation_height"] = -45.0
        prob["monopile_height"] = hval.sum()
        prob["monopile_outer_diameter_in"] = 10.0 * np.ones(3)
        prob["monopile_top_diameter"] = 10.0
        prob["monopile_layer_thickness"] = 0.1 * np.ones(3).reshape((1, 3))
        prob["outfitting_factor_in"] = 1.0
        prob["monopile_layer_materials"] = prob["monopile_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8
        prob["sigma_ult_mat"] = 1e8 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 1e1
        prob["wohler_A_mat"] = 1e1

        prob["yaw"] = 0.0
        prob["wind_reference_height"] = 80.0
        prob["z0"] = 0.0
        prob["cd_usr"] = -1.0
        prob["rho_air"] = 1.225
        prob["mu_air"] = 1.7934e-5
        prob["shearExp"] = 0.2
        prob["rho_water"] = 1025.0
        prob["mu_water"] = 1.3351e-3
        prob["beta_wind"] = prob["beta_wave"] = 0.0
        prob["Hsig_wave"] = 0.0
        prob["Tsig_wave"] = 1e3
        prob["env.Uref"] = 15.0
        prob["turbine_F"] = 1e3 * np.arange(2, 5)
        prob["turbine_M"] = 1e4 * np.arange(2, 5)
        prob["turbine_mass"] = 1e5
        prob["turbine_cg"] = 1e1 * np.ones(3)
        prob["turbine_I"] = 2e1 * np.ones(6)
        prob.run_model()

        # All other tests from above
        mass_dens = 1e4 * (5.0**2 - 4.9**2) * np.pi
        npt.assert_equal(prob["z_start"], -45.0)
        npt.assert_equal(prob["transition_piece_height"], 0.0)
        npt.assert_equal(prob["suctionpile_depth"], 15.0)
        npt.assert_equal(prob["z_param"], np.array([-45.0, -30.0, 0.0]))
        npt.assert_almost_equal(prob["monopile_mass"], mass_dens * 45.0 + 1e2 + 1e4)

        npt.assert_equal(prob.model.perf.monopile.frame.rnode, np.array([1], dtype=np.int_))
        npt.assert_equal(prob.model.perf.monopile.frame.rKx, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.monopile.frame.rKy, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.monopile.frame.rKz, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.monopile.frame.rKtx, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.monopile.frame.rKty, np.array([RIGID]))
        npt.assert_equal(prob.model.perf.monopile.frame.rKtz, np.array([RIGID]))

        npt.assert_equal(prob.model.perf.monopile.frame.ENMnode, np.array([6, 5, 1], dtype=np.int_))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMmass, np.array([1e5, 1e2, 1e4]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhox, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhoy, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMrhoz, np.r_[10.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxx, np.array([20, 1e2 * 25 * 0.5, 1e4 * 25 * 0.25]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIyy, np.array([20, 1e2 * 25 * 0.5, 1e4 * 25 * 0.25]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIzz, np.array([20, 1e2 * 25, 1e4 * 25 * 0.5]))
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxy, np.r_[20.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIxz, np.r_[20.0, 0, 0])
        npt.assert_equal(prob.model.perf.monopile.frame.ENMIyz, np.r_[20.0, 0, 0])
        self.assertEqual(prob.model.perf.monopile.frame.addGravityLoadForExtraNodeMass, [False, True, True])

        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].NF, np.array([6]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fx, np.array([2e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fy, np.array([3e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Fz, np.array([4e3]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Mxx, np.array([2e4]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Myy, np.array([3e4]))
        npt.assert_equal(prob.model.perf.monopile.frame.loadCases[0].Mzz, np.array([4e4]))

        # npt.assert_almost_equal(prob["monopile.mudline_F"], [3.74393291e04, 1.84264671e03, -3.39826364e07], 0)
        # npt.assert_almost_equal(prob["monopile.mudline_M"], [-294477.83027742, -2732413.3684215, 40000.0], 0)

    def testAddedMassForces(self):
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = 3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["FixedBottomSE"]["soil_springs"] = False
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gravity_foundation"] = False

        prob = om.Problem(reports=False)
        prob.model = mon.MonopileSE(modeling_options=self.modeling_options)
        prob.setup()

        prob["water_depth"] = 30.0
        prob["transition_piece_mass"] = 0.0
        prob["transition_piece_cost"] = 0.0
        prob["gravity_foundation_mass"] = 0.0

        hval = np.array([15.0, 30.0])
        prob["monopile_s"] = np.cumsum(np.r_[0, hval]) / hval.sum()
        prob["monopile_foundation_height"] = -45.0
        prob["monopile_height"] = hval.sum()
        prob["monopile_outer_diameter_in"] = 10.0 * np.ones(3)
        prob["monopile_top_diameter"] = 10.0
        prob["monopile_layer_thickness"] = 0.1 * np.ones((1, 3))
        prob["outfitting_factor_in"] = 1.0
        prob["monopile_layer_materials"] = prob["monopile_layer_materials"] = ["steel"]
        prob["material_names"] = ["steel"]
        prob["E_mat"] = 1e9 * np.ones((1, 3))
        prob["G_mat"] = 1e8 * np.ones((1, 3))
        prob["rho_mat"] = 1e4
        prob["sigma_y_mat"] = 1e8
        prob["sigma_ult_mat"] = 1e8 * np.ones((1, 3))
        prob["wohler_exp_mat"] = 1e1
        prob["wohler_A_mat"] = 1e1

        prob["yaw"] = 0.0
        # prob["G_soil"] = 1e7
        # prob["nu_soil"] = 0.5
        prob["wind_reference_height"] = 80.0
        prob["z0"] = 0.0
        prob["ca_usr"] = 0.0
        prob["cd_usr"] = -1.0
        prob["rho_air"] = 1.225
        prob["mu_air"] = 1.7934e-5
        prob["shearExp"] = 0.2
        prob["rho_water"] = 1025.0
        prob["mu_water"] = 1.3351e-3
        prob["beta_wind"] = prob["beta_wave"] = 0.0
        prob["Hsig_wave"] = 0.0
        prob["Tsig_wave"] = 1e3
        prob["env.Uref"] = 15.0
        prob["turbine_F"] = 1e3 * np.arange(2, 5)
        prob["turbine_M"] = 1e4 * np.arange(2, 5)
        prob.run_model()

        myFz = copy.copy(prob["monopile.monopile_Fz"])

        prob["transition_piece_mass"] = 1e2
        prob.run_model()
        myFz[3:-2] -= 1e2 * g
        npt.assert_almost_equal(prob["monopile.monopile_Fz"], myFz)

        prob["gravity_foundation_mass"] = 1e3
        prob.run_model()
        # myFz[0] -= 1e3*g
        npt.assert_almost_equal(prob["monopile.monopile_Fz"], myFz)

    def testExampleRegression(self):
        # --- geometry ----
        h_param = np.diff(np.array([-60.0, -30.0, 10.0]))
        d_param = np.array([6.0, 6.0, 5.5])
        t_param = 1.3 * np.array([0.027, 0.023, 0.019])
        z_foundation = -60.0
        yaw = 0.0
        Koutfitting = 1.07

        # --- material props ---
        E = 210e9
        G = 80.8e9
        rho = 8500.0
        sigma_y = 450.0e6
        sigma_ult = wohler_A = 500e6
        wohler_exp = 4.0

        # --- wind ---
        wind_zref = 90.0
        wind_z0 = 0.0
        shearExp = 0.2
        cd_usr = -1.0
        # ---------------

        # --- wave ---
        water_depth = 30.0
        soilG = 140e6
        soilnu = 0.4
        Hsig = 4.0
        Tsig = 10.0
        # ---------------

        # --- costs ---
        material_cost = 5.0
        labor_cost = 100.0 / 60.0
        painting_cost = 30.0
        # ---------------

        # two load cases.  TODO: use a case iterator

        # # --- loading case 1: max Thrust ---
        wind_Uref1 = 11.73732
        Fx1 = 1284744.19620519
        Fy1 = 0.0
        Fz1 = -2914124.84400512
        Mxx1 = 3963732.76208099
        Myy1 = -2275104.79420872
        Mzz1 = -346781.68192839
        # # ---------------

        # # --- loading case 2: max wind speed ---
        wind_Uref2 = 70.0
        Fx2 = 930198.60063279
        Fy2 = 0.0
        Fz2 = -2883106.12368949
        Mxx2 = -1683669.22411597
        Myy2 = -2522475.34625363
        Mzz2 = 147301.97023764
        # # ---------------

        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = len(d_param)
        self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = 1
        self.modeling_options["WISDEM"]["n_dlc"] = 2
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_f"] = 1.35
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_m"] = 1.3
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_n"] = 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_b"] = 1.1
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gamma_fatigue"] = 1.35 * 1.3 * 1.0
        self.modeling_options["WISDEM"]["FixedBottomSE"]["soil_springs"] = True
        self.modeling_options["WISDEM"]["FixedBottomSE"]["gravity_foundation"] = False

        def fill_prob():
            prob = om.Problem(reports=False)
            prob.model = mon.MonopileSE(modeling_options=self.modeling_options)
            prob.setup()

            if self.modeling_options["WISDEM"]["FixedBottomSE"]["wind"] == "PowerWind":
                prob["shearExp"] = shearExp

            # assign values to params

            # --- geometry ----
            prob["water_depth"] = water_depth
            prob["monopile_s"] = np.cumsum(np.r_[0.0, h_param]) / h_param.sum()
            prob["monopile_foundation_height"] = z_foundation
            prob["tower_foundation_height"] = z_foundation + h_param.sum()
            prob["monopile_height"] = h_param.sum()
            prob["monopile_outer_diameter_in"] = d_param
            prob["monopile_top_diameter"] = d_param[-1]
            prob["monopile_layer_thickness"] = t_param.reshape((1, len(t_param)))
            prob["outfitting_factor_in"] = Koutfitting
            prob["monopile_layer_materials"] = ["steel"]
            prob["material_names"] = ["steel"]
            prob["yaw"] = yaw
            prob["G_soil"] = soilG
            prob["nu_soil"] = soilnu
            prob["Hsig_wave"] = Hsig
            prob["Tsig_wave"] = Tsig
            # --- material props ---
            prob["E_mat"] = E * np.ones((1, 3))
            prob["G_mat"] = G * np.ones((1, 3))
            prob["rho_mat"] = rho
            prob["sigma_y_mat"] = sigma_y
            prob["sigma_ult_mat"] = sigma_ult * np.ones((1, 3))
            prob["wohler_exp_mat"] = wohler_exp
            prob["wohler_A_mat"] = wohler_A

            # --- costs ---
            prob["unit_cost_mat"] = [material_cost]
            prob["labor_cost_rate"] = labor_cost
            prob["painting_cost_rate"] = painting_cost
            # -----------

            # --- wind & wave ---
            prob["wind_reference_height"] = wind_zref
            prob["z0"] = wind_z0
            prob["cd_usr"] = cd_usr
            prob["ca_usr"] = 0.0 # Default was zero before, but now the default is -1
            prob["rho_air"] = 1.225
            prob["mu_air"] = 1.7934e-5

            # # --- loading case 1: max Thrust ---
            # # --- loading case 2: max Wind Speed ---
            prob["env1.Uref"] = wind_Uref1
            prob["env2.Uref"] = wind_Uref2

            prob["turbine_F"] = np.c_[[Fx1, Fy1, Fz1], [Fx2, Fy2, Fz2]]
            prob["turbine_M"] = np.c_[[Mxx1, Myy1, Mzz1], [Mxx2, Myy2, Mzz2]]

            return prob

        # # --- run ---
        prob = fill_prob()
        prob.run_model()

        npt.assert_almost_equal(prob["z_full"], [0.0, 10.0, 20.0, 30.0, 43.33333333, 56.66666667, 70.0])
        npt.assert_almost_equal(prob["outer_diameter_full"], [6.0, 6.0, 6.0, 6.0, 5.833333333, 5.6666667, 5.5])
        npt.assert_almost_equal(prob["t_full"], [0.0325, 0.0325, 0.0325, 0.0273, 0.0273, 0.0273])

        npt.assert_almost_equal(prob["monopile_mass"], [344802.1109282])
        npt.assert_almost_equal(prob["monopile_z_cg"], [32.9740276])
        npt.assert_almost_equal(prob["constr_d_to_t"], [184.6153846, 210.6227106])
        npt.assert_almost_equal(prob["constr_taper"], [1.0, 0.9166667])
        npt.assert_almost_equal(prob["env1.Uref"], [11.73732])
        npt.assert_almost_equal(prob["env2.Uref"], [70.0])
        npt.assert_almost_equal(prob["f1"], [2.8398576087517053], 3)
        npt.assert_almost_equal(prob["monopile.top_deflection"], [0.04355821, 0.02954024], 3)
        npt.assert_almost_equal(
            prob["post.constr_stress"].T,
            [
                [0.03385147, 0.01624988, 0.11893221, 0.13589145, 0.05543162, 0.0103132 ],
                [0.0341598 , 0.01988708, 0.08923635, 0.11635987, 0.04822173, 0.01055353],
           ],
            3,
        )
        npt.assert_almost_equal(
            prob["post.constr_global_buckling"].T,
            [
                [0.11201532, 0.12416107, 0.19921349, 0.17590641, 0.12969364, 0.08627106],
                [0.11175936, 0.11925156, 0.17013531, 0.15228978, 0.12254481, 0.08627106],
            ],
            1,
        )
        npt.assert_almost_equal(
            prob["post.constr_shell_buckling"].T,
            [
                [3.08691592e-03, 1.15049491e-03, 5.89397300e-01, 1.76100515e+00, 6.40018188e-01, 8.86263582e-02],
                [3.13317242e-03, 1.36025565e-03, 5.79068840e-01, 1.75232887e+00, 6.36781099e-01, 9.12750704e-02],
            ],
            3,
        )
        npt.assert_almost_equal(prob["monopile.mudline_F"][0, :], [1284980.175382  ,  930363.07603095 ], 2)
        npt.assert_array_less(np.abs(prob["monopile.mudline_F"][1, :]), 1e2, 2)
        npt.assert_almost_equal(prob["monopile.mudline_F"][2, :], [-4732577.31059091, -4735037.20580716], 2)
        npt.assert_almost_equal(
            prob["monopile.mudline_M"].T,
            [[3983126.23029407, 32107268.56484913,  -346819.26780544], [-1691961.71385695, 22367653.91976971,   147317.94383742]],
            0,
        )

        # Now regression on DNV-GL C202 methods
        self.modeling_options["WISDEM"]["FixedBottomSE"]["buckling_method"] = "dnvgl"
        prob = fill_prob()
        prob.run_model()

        npt.assert_almost_equal(prob["z_full"], [0.0, 10.0, 20.0, 30.0, 43.33333333, 56.66666667, 70.0])
        npt.assert_almost_equal(prob["outer_diameter_full"], [6.0, 6.0, 6.0, 6.0, 5.833333333, 5.6666667, 5.5])
        npt.assert_almost_equal(prob["t_full"], [0.0325, 0.0325, 0.0325, 0.0273, 0.0273, 0.0273])

        npt.assert_almost_equal(prob["monopile_mass"], [344802.1109282])
        npt.assert_almost_equal(prob["monopile_z_cg"], [32.9740276])
        npt.assert_almost_equal(prob["constr_d_to_t"], [184.6153846, 210.6227106])
        npt.assert_almost_equal(prob["constr_taper"], [1.0, 0.9166667])
        npt.assert_almost_equal(prob["env1.Uref"], [11.73732])
        npt.assert_almost_equal(prob["env2.Uref"], [70.0])
        npt.assert_almost_equal(prob["f1"], [2.83985761], 3)
        npt.assert_almost_equal(prob["monopile.top_deflection"], [0.04355821, 0.02954024], 3)
        npt.assert_almost_equal(
            prob["post.constr_stress"].T,
            [
                [0.03385147, 0.01624988, 0.11893221, 0.13589145, 0.05543162, 0.0103132 ],
                [0.0341598 , 0.01988708, 0.08923635, 0.11635987, 0.04822173, 0.01055353],
            ],
            2,
        )
        npt.assert_almost_equal(
            prob["post.constr_global_buckling"].T,
            [
                [3.29521915e-02, 3.70069327e-02, 6.99117262e-02, 5.96692606e-02, 3.61803070e-02, 1.02510833e-10],
                [3.27627239e-02, 3.33331034e-02, 4.83927465e-02, 4.33297244e-02, 3.11606296e-02, 5.44601822e-11],
            ],
            1,
        )
        npt.assert_almost_equal(
            prob["post.constr_shell_buckling"].T,
            [
                [0.03960002, 0.04474592, 0.57096438, 1.74890517, 0.8283988 , 0.16249855],
                [0.03919867, 0.04176198, 0.56125835, 1.73421614, 0.81345437, 0.16628092],
            ],
            2,
        )
        npt.assert_almost_equal(prob["monopile.mudline_F"][0, :], [1284980.175382  ,  930363.07603095], 2)
        npt.assert_array_less(np.abs(prob["monopile.mudline_F"][1, :]), 1e2, 2)
        npt.assert_almost_equal(prob["monopile.mudline_F"][2, :], [-4732577.31, -4735037.21], 2)
        npt.assert_almost_equal(
            prob["monopile.mudline_M"].T,
            [[3983126.23029407, 32107268.56484913,  -346819.26780544], [-1691961.71385695, 22367653.91976971,   147317.94383742]],
            0,
        )


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestPreDiscretization),
        unittest.TestLoader().loadTestsFromTestCase(TestMass),
        unittest.TestLoader().loadTestsFromTestCase(TestMonopileSE),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import numpy as np
import openmdao.api as om

import wisdem.commonse.fileIO as fio
from wisdem.drivetrainse.generator import Generator

opt_flag = False
n_pc = 20

# Example optimization of a generator for costs on a 5 MW reference turbine
prob = om.Problem(reports=False)
prob.model = Generator(design="eesg", n_pc=n_pc)

if opt_flag:
    # add optimizer and set-up problem (using user defined input on objective function)
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"

    # Specificiency target efficiency(%)
    Eta_Target = 0.930

    eps = 1e-6

    # Set up design variables and bounds for a SCIG designed for a 5MW turbine
    # Design variables
    prob.model.add_design_var("r_s", lower=0.5, upper=9.0)
    prob.model.add_design_var("len_s", lower=0.5, upper=2.5)
    prob.model.add_design_var("h_s", lower=0.06, upper=0.15)
    prob.model.add_design_var("tau_p", lower=0.04, upper=0.2)
    prob.model.add_design_var("N_f", lower=10, upper=300, ref=1e2)
    prob.model.add_design_var("I_f", lower=1, upper=500, ref=1e2)
    prob.model.add_design_var("n_r", lower=5.0, upper=15.0)
    prob.model.add_design_var("h_yr", lower=0.01, upper=0.25)
    prob.model.add_design_var("h_ys", lower=0.01, upper=0.25)
    prob.model.add_design_var("b_arm", lower=0.1, upper=1.5)
    prob.model.add_design_var("d_r", lower=0.1, upper=1.5)
    prob.model.add_design_var("t_wr", lower=0.001, upper=0.2)
    prob.model.add_design_var("n_s", lower=5.0, upper=15.0)
    prob.model.add_design_var("b_st", lower=0.1, upper=1.5)
    prob.model.add_design_var("d_s", lower=0.1, upper=1.5)
    prob.model.add_design_var("t_ws", lower=0.001, upper=0.2)

    # Constraints
    prob.model.add_constraint("B_gfm", lower=0.617031, upper=1.057768)
    prob.model.add_constraint("B_pc", upper=2.0)
    prob.model.add_constraint("E_s", lower=500.0, upper=5000.0, ref=1e3)
    prob.model.add_constraint("J_f", upper=6.0)
    prob.model.add_constraint("n_brushes", upper=6)
    prob.model.add_constraint("Power_ratio", upper=2 - eps)
    prob.model.add_constraint("B_symax", upper=2.0 - eps)
    prob.model.add_constraint("B_rymax", upper=2.0 - eps)
    prob.model.add_constraint("B_tmax", upper=2.0 - eps)
    prob.model.add_constraint("B_g", lower=0.7, upper=1.2)
    prob.model.add_constraint("con_uas", lower=0.0 + eps)
    prob.model.add_constraint("con_zas", lower=0.0 + eps)
    prob.model.add_constraint("con_yas", lower=0.0 + eps)
    prob.model.add_constraint("con_uar", lower=0.0 + eps)
    prob.model.add_constraint("con_yar", lower=0.0 + eps)
    prob.model.add_constraint("con_TC2r", lower=0.0 + eps)
    prob.model.add_constraint("con_TC2s", lower=0.0 + eps)
    prob.model.add_constraint("con_bst", lower=0.0 - eps)
    prob.model.add_constraint("A_1", upper=60000.0 - eps, ref=1e5, indices=[-1])
    prob.model.add_constraint("J_s", upper=6.0)
    prob.model.add_constraint("A_Cuscalc", lower=5.0, upper=300, ref=1e2)
    prob.model.add_constraint("K_rad", lower=0.2 + eps, upper=0.27)
    prob.model.add_constraint("Slot_aspect_ratio", lower=4.0, upper=10.0)
    prob.model.add_constraint("generator_efficiency", lower=Eta_Target, indices=[-1])
    prob.model.add_constraint("con_zar", lower=0.0 + eps)
    prob.model.add_constraint("con_br", lower=0.0 + eps)

    Objective_function = "generator_cost"
    prob.model.add_objective(Objective_function, scaler=1e-5)


prob.setup()

# Specify Target machine parameters
prob["machine_rating"] = 5000000.0
prob["rated_torque"] = 4.143289e6
prob["shaft_rpm"] = np.linspace(2, 12.1, n_pc)  # 8.68                # rpm 9.6
prob["sigma"] = 48.373e3
# Initial design variables
# prob['r_s']     = 3.2
prob["rad_ag"] = 3.2
prob["len_s"] = 1.4
prob["h_s"] = 0.060
prob["tau_p"] = 0.170
prob["I_f"] = 69
prob["N_f"] = 100
prob["h_ys"] = 0.130
prob["h_yr"] = 0.120
prob["n_s"] = 5
prob["b_st"] = 0.470
prob["n_r"] = 5
prob["b_r"] = 0.480
prob["d_r"] = 0.510
prob["d_s"] = 0.400
prob["t_wr"] = 0.140
prob["t_ws"] = 0.070
prob["D_shaft"] = 2 * 0.43  # 10MW: 0.523950817,#5MW: 0.43, #3MW:0.363882632 #1.5MW: 0.2775  0.75MW: 0.17625
prob["q1"] = 2
# ---------------------------------------------------
prob["E"] = 2e11
prob["G"] = 79.3e9

# Specific costs
prob["C_Cu"] = 4.786  # Unit cost of Copper $/kg
prob["C_Fe"] = 0.556  # Unit cost of Iron $/kg
prob["C_Fes"] = 0.50139  # specific cost of Structural_mass $/kg
prob["C_PM"] = 95.0

# Material properties
prob["rho_Fe"] = 7700.0  # Steel density Kg/m3
prob["rho_Fes"] = 7850  # structural Steel density Kg/m3
prob["rho_Copper"] = 8900.0  # copper density Kg/m3
prob["rho_PM"] = 7450.0  # typical density Kg/m3 of neodymium magnets (added 2019 09 18) - for pmsg_[disc|arms]

# Run optimization
if opt_flag:
    prob.model.approx_totals()
    prob.run_driver()
else:
    prob.run_model()

prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)
# fio.save_data('EESG', prob, npz_file=False, mat_file=False, xls_file=True)
print("Efficiency table:")
print(" rpm     converter  transformer  generator   ")
print("-------------------------------------------")
print(np.c_[prob["shaft_rpm"],
            prob["converter_efficiency"],
            prob["transformer_efficiency"],
            prob["generator_efficiency"],
            ])



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


from copy import deepcopy
from itertools import product

import pytest

from wisdem.orbit.phases.design import OffshoreSubstationDesign

base = {
    "site": {"depth": 30},
    "plant": {"num_turbines": 50},
    "turbine": {"turbine_rating": 6},
    "substation_design": {},
}


@pytest.mark.parametrize(
    "depth,num_turbines,turbine_rating",
    product(range(10, 51, 10), range(3, 13, 1), range(20, 80, 10)),
)
def test_parameter_sweep(depth, num_turbines, turbine_rating):
    config = {
        "site": {"depth": depth},
        "plant": {"num_turbines": num_turbines},
        "turbine": {"turbine_rating": turbine_rating},
        "substation_design": {},
    }

    o = OffshoreSubstationDesign(config)
    o.run()

    # Check valid substructure length
    assert 10 <= o._outputs["offshore_substation_substructure"]["length"] <= 80

    # Check valid substructure mass
    assert 200 <= o._outputs["offshore_substation_substructure"]["mass"] <= 2500

    # Check valid topside mass
    assert 200 <= o._outputs["offshore_substation_topside"]["mass"] <= 5000

    # Check valid substation cost
    assert 1e6 <= o.total_cost <= 300e6


def test_oss_kwargs():
    test_kwargs = {
        "mpt_cost_rate": 13500,
        "topside_fab_cost_rate": 15500,
        "topside_design_cost": 5e6,
        "shunt_cost_rate": 40000,
        "switchgear_cost": 15e5,
        "backup_gen_cost": 2e6,
        "workspace_cost": 3e6,
        "other_ancillary_cost": 4e6,
        "topside_assembly_factor": 0.08,
        "oss_substructure_cost_rate": 7250,
        "oss_pile_cost_rate": 2500,
        "num_substations": 2,
    }

    o = OffshoreSubstationDesign(base)
    o.run()
    base_cost = o.total_cost

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["substation_design"] = {}
        config["substation_design"][k] = v

        o = OffshoreSubstationDesign(config)
        o.run()
        cost = o.total_cost

        assert cost != base_cost



import numpy as np
import math
from wisdem.landbosse.model.WeatherDelay import WeatherDelay as WD
import traceback
from wisdem.landbosse.model.CostModule import CostModule
import pandas as pd

class SitePreparationCost(CostModule):
    """
    **SitePreparationCost.py**

    - Created by Annika Eberle and Owen Roberts on Apr. 3, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on Jun 3, 2019



    **Calculates cost of constructing roads for land-based wind projects:**

    - Get terrain complexity

    - Get turbine spacing

    - Get road width

    - Get number of turbines

    - Get turbine rating

    - Get duration of construction*  #todo: add to process diagram

    - Get road length

    - Get weather data

    - Get road thickness

    - Calculate volume of road based on road thickness, road width, and road length

    - Calculate road labor and equipment costs by operation and type using RSMeans data

    - Calculate man hours and equipment hours for compaction of soil based on road length, road thickness, soil type, road width, and equipment size

    - Calculate man hours and equipment hours for mass material movement based on land cover, terrain complexity, and road length

    - Calculate man hours and equipment hours for rock placement based on equipment size, distance to quarry, and volume of road

    - Calculate man hours and equipment hours for compaction of rock based on road length, road thickness, and rock type

    - Calculate man hours and equipment hours for final grading based on road length

    - Calculate quantity of materials based on volume of materials

    - Calculate material costs by type

    - Calculate material costs using quantity of materials by material type and material prices by material type

    - Sum road costs over all operations and material types to get total costs by type of cost (e.g., material vs. equipment)

    - Return total road costs by type of cost


    **Keys in the input dictionary are the following:**

    - road_length

    - road_width

    - road_thickness

    - crane_width

    - num_access_roads  #TODO: Add to excel inputs sheet

    - num_turbines

    - rsmeans (dataframe)

    - duration_construction

    - wind_delays

    - wind_delay_time

    - material_price (dataframe)

    - rsmeans_per_diem

    - rotor_diameter_m

    - turbine_spacing_rotor_diameters


    **Keys in the output dictionary are the following:**

    - road_length_m

    - road_volume_m3

    - depth_to_subgrade_m

    - crane_path_widthm

    - road_thickess_m

    - road_width_m

    - road_width_m

    - material_volume_cubic_yards

    - road_construction_time

    - topsoil_volume

    - embankment_volume_crane

    - embankment_volume_road

    - rough_grading_area

    - material_needs (dataframe)

    - operation_data (dataframe)

    - total_road_cost



    """


    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

        #Road quality and fraction of roads that are
        self.fraction_new_roads = self.input_dict['fraction_new_roads']
        self.road_quality  = self.input_dict['road_quality']

        # Conversion factors. Making this data private (hidden from outside of this class):
        self._meters_per_foot = 0.3
        self._meters_per_inch = 0.025
        self._cubic_yards_per_cubic_meter = 1.30795
        self._square_feet_per_square_meter = 10.7639

        # cubic meters for crane pad and maintenance ring for each turbine
        # (from old BOS model - AI - Access Roads & Site Imp. tab cell J33)
        self._crane_pad_volume = 125

        # conversion factor for converting packed cubic yards to loose cubic yards
        # material volume is about 1.4 times greater than road volume due to compaction
        self._yards_loose_per_yards_packed = 1.39

        # calculated road properties
        self._lift_depth_m = 0.2

    def calculate_road_properties(self, road_properties_input, road_properties_output):
        """
        Calculates the volume of road materials need based on length, width, and thickness of roads

        Parameters
        ----------

        int num_turbines
            number of turbines in wind farm

        unitless turbine_spacing_rotor_diameters
            Immediate spacing between two turbines as a function of rotor diameter

        int rotor_diameter_m
            Rotor diameter

        float crane_width
            Crane width in meters

        float road_thickness
            Road thickness in inches

        float road_width_ft
            Road width in feet



        Returns
        ----------

        road_length_m
            Calculated road length (in m)

        road_volume_m3
            Calculated road volume (in m^3)

        material_volume_cubic_yards
            Material volume required (in cubic yards) based on road volume


        """

        if road_properties_input['road_distributed_wind'] == True or road_properties_input['turbine_rating_MW'] < 0.1:
            road_properties_output['road_volume'] = road_properties_input['road_length_adder_m'] * \
                                                    (road_properties_input['road_width_ft'] * self._meters_per_foot) * \
                                                    (road_properties_input[
                                                         'road_thickness'] * self._meters_per_inch)  # units of cubic meters
        else:
            road_properties_output['road_length_m'] = ((road_properties_input['num_turbines'] - 1) *
                                                       road_properties_input['turbine_spacing_rotor_diameters'] *
                                                       road_properties_input['rotor_diameter_m']) + \
                                                      road_properties_input['road_length_adder_m']
            road_properties_output['road_width_m'] = road_properties_input['road_width_ft'] * self._meters_per_foot
            road_properties_output['road_volume'] = road_properties_output['road_length_m'] * \
                                                    (road_properties_input['road_width_ft'] * self._meters_per_foot) * \
                                                    (road_properties_input[
                                                         'road_thickness'] * self._meters_per_inch)  # units of cubic meters


        road_properties_output['road_volume_m3'] = road_properties_output['road_volume'] + self._crane_pad_volume * \
                                                   road_properties_input['num_turbines']  # in cubic meters

        road_properties_output['depth_to_subgrade_m'] = 0.1

        road_properties_output['crane_path_width_m'] = road_properties_input['crane_width'] + 1.5

        # todo: replace with actual crane path width from erection module
        road_properties_output['road_thickness_m'] = (road_properties_input['road_thickness'] * self._meters_per_inch)
        road_properties_output['road_width_m'] = road_properties_input['road_width_ft'] * self._meters_per_foot
        road_properties_output['material_volume_cubic_yards'] = (road_properties_output['road_volume_m3'] * self._cubic_yards_per_cubic_meter * self._yards_loose_per_yards_packed)  # todo: output_dict material volume

        return road_properties_output

    def estimate_construction_time(self, estimate_construction_time_input, estimate_construction_time_output):
        """
        Estimates construction time of roads for entire project.

        Parameters
        ----------
        float crane_path_width_m
            Width of crane path (in m)

        float road_length_m
            Road length (in m)

        float depth_to_subgrade_m
            Depth to subgarde (in m)

        float road_volume
            Road volume (in m^3)

        float road_thickness_m
            Road thickness (in m)



        Returns
        ----------

        pd.DataFrame operation_data
            Dataframe which conatains following outputs:

            -  Number of days required for construction

            - Number of crews required to complete roads construction in specified construction time

            - Cost of labor and equipment rental prior to weather delays

        """
        throughput_operations = estimate_construction_time_input['rsmeans']

        #TODO: Figure out where 'construct_duration' gets read in.
        estimate_construction_time_output['road_construction_time'] = estimate_construction_time_input[
                                                                          'construct_duration'] * 1 / 5  # assumes road construction occurs for 1/5 of project time

        # Main switch between small DW wind and (utility scale + distributed wind)
        # select operations for roads module that have data
        if estimate_construction_time_input['turbine_rating_MW'] >= 0.1:
            operation_data = throughput_operations.where(throughput_operations['Module'] == 'Roads').dropna(thresh=4)
        else:
            operation_data = throughput_operations.where(throughput_operations['Module'] == 'Small DW Roads').dropna(
                thresh=4)
            operation_data = operation_data.dropna(subset=['Units'])

        # create list of unique material units for operations
        list_units = operation_data['Units'].unique()

        if estimate_construction_time_input['road_distributed_wind'] == True and estimate_construction_time_input[
            'turbine_rating_MW'] >= 0.1:
            estimate_construction_time_output['topsoil_volume'] = estimate_construction_time_input[
                                                                      'site_prep_area_m2'] * (
                                                                  estimate_construction_time_output[
                                                                      'depth_to_subgrade_m']) * self._cubic_yards_per_cubic_meter  # units: cubic yards
            estimate_construction_time_output['embankment_volume_crane'] = estimate_construction_time_output[
                'topsoil_volume']  # units: cubic yards
            estimate_construction_time_output['embankment_volume_road'] = estimate_construction_time_output[
                'topsoil_volume']  # units: cubic yards
            estimate_construction_time_output['rough_grading_area'] = (estimate_construction_time_input[
                                                                           'site_prep_area_m2'] * 10.76391) / 100000  # where 10.76391 sq ft = 1 sq m

        elif estimate_construction_time_input['road_distributed_wind'] == True and estimate_construction_time_input[
            'turbine_rating_MW'] < 0.1:
            estimate_construction_time_output['topsoil_volume'] = estimate_construction_time_input[
                                                                      'road_length_adder_m'] * (
                                                                              estimate_construction_time_input[
                                                                                  'road_width_ft'] * 0.3048) * (
                                                                              estimate_construction_time_input[
                                                                                  'road_thickness'] * 0.0254) * 1.308  # Units: CY (where 1 m3 = 1.308 CY)

        else:
            estimate_construction_time_output['topsoil_volume'] = (estimate_construction_time_output[
                'crane_path_width_m']) * estimate_construction_time_output['road_length_m'] * (
                                                                  estimate_construction_time_output[
                                                                      'depth_to_subgrade_m']) * self._cubic_yards_per_cubic_meter  # units: cubic yards
            estimate_construction_time_output['embankment_volume_crane'] = (estimate_construction_time_output[
                'crane_path_width_m']) * estimate_construction_time_output['road_length_m'] * (
                                                                           estimate_construction_time_output[
                                                                               'depth_to_subgrade_m']) * self._cubic_yards_per_cubic_meter  # units: cubic yards
            estimate_construction_time_output['embankment_volume_road'] = estimate_construction_time_output[
                                                                              'road_volume'] * self._cubic_yards_per_cubic_meter * math.ceil(
                estimate_construction_time_output['road_thickness_m'] / self._lift_depth_m)  # units: cubic yards road
            estimate_construction_time_output['rough_grading_area'] = estimate_construction_time_output[
                                                                          'road_length_m'] * \
                                                                      estimate_construction_time_output[
                                                                          'road_width_m'] * \
                                                                      self._square_feet_per_square_meter * \
                                                                      math.ceil(estimate_construction_time_output[
                                                                                    'road_thickness_m'] / self._lift_depth_m) / 100000  # Units: Each (100,000 square feet)

        if estimate_construction_time_input['turbine_rating_MW'] >= 0.1:
            material_quantity_dict = {'cubic yard': estimate_construction_time_output['topsoil_volume'],
                                      'embankment cubic yards crane': estimate_construction_time_output[
                                          'embankment_volume_crane'],
                                      'embankment cubic yards road': estimate_construction_time_output[
                                          'embankment_volume_road'],
                                      'loose cubic yard': estimate_construction_time_output[
                                          'material_volume_cubic_yards'],
                                      'Each (100000 square feet)': estimate_construction_time_output[
                                          'rough_grading_area']}
        else:  # small DW
            material_quantity_dict = {'cubic yard': estimate_construction_time_output['topsoil_volume'],
                                      'embankment cubic yards crane': estimate_construction_time_output[
                                          'topsoil_volume'],
                                      'loose cubic yard': estimate_construction_time_output['topsoil_volume'],
                                      'embankment cubic yards road': estimate_construction_time_output['topsoil_volume']
                                      }

        material_needs = pd.DataFrame(columns=['Units', 'Quantity of material'])
        for unit in list_units:
            unit_quantity = pd.DataFrame([[unit, material_quantity_dict[unit]]],
                                         columns=['Units', 'Quantity of material'])
            material_needs = pd.concat((material_needs, unit_quantity))

        estimate_construction_time_output['material_needs'] = material_needs

        # join material needs with operational data to compute costs
        operation_data = pd.merge(operation_data, material_needs, on=['Units'], sort=True).dropna(thresh=3)
        operation_data = operation_data.where((operation_data['Daily output']).isnull() == False).dropna(thresh=4)

        # calculate operational parameters and estimate costs without weather delays
        operation_data['Number of days'] = operation_data['Quantity of material'] / operation_data['Daily output']
        operation_data['Number of crews'] = np.ceil((operation_data['Number of days'] / 30.) / estimate_construction_time_output['road_construction_time'])
        operation_data['Cost USD without weather delays'] = operation_data['Quantity of material'] * operation_data['Rate USD per unit']

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that time frame for weather delays; if not, use the number of days calculated
        operation_data['time_construct_bool'] = operation_data['Number of days'] > estimate_construction_time_output['road_construction_time'] * 30.
        boolean_dictionary = {True: estimate_construction_time_output['road_construction_time'] * 30., False: np.NAN}
        operation_data['time_construct_bool'] = operation_data['time_construct_bool'].map(boolean_dictionary)
        operation_data['Time construct days'] = operation_data[['time_construct_bool', 'Number of days']].min(axis=1)
        num_days = operation_data['Time construct days'].max()

        # pull out management data
        if estimate_construction_time_input['turbine_rating_MW'] >= 0.1:
            crew_cost = self.input_dict['crew_cost']
            crew = self.input_dict['crew'][self.input_dict['crew']['Crew type ID'].str.contains('M0')]
            management_crew = pd.merge(crew_cost, crew, on=['Labor type ID'], sort=True)
            management_crew = management_crew.assign(per_diem_total=
                                                     management_crew['Per diem USD per day'] *
                                                     management_crew['Number of workers'] *
                                                     num_days)
            management_crew = management_crew.assign(hourly_costs_total=
                                                     management_crew['Hourly rate USD per hour'] *
                                                     self.input_dict['hour_day'][self.input_dict['time_construct']] *
                                                     num_days)
            management_crew = management_crew.assign(total_crew_cost_before_wind_delay=
                                                     management_crew['per_diem_total'] +
                                                     management_crew['hourly_costs_total'])
            self.output_dict['management_crew'] = management_crew

            self.output_dict['managament_crew_cost_before_wind_delay'] = management_crew[
                                                     'total_crew_cost_before_wind_delay'].sum()

        estimate_construction_time_output['operation_data'] = operation_data

        return operation_data

    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """
        Calculates wind delay for roads.

        """

        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data['wind_delays'])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data['wind_delay_time'] = float(wind_delay.sum().iloc[0])

        return weather_delay_output_data


    def new_and_existing_total_road_cost(self, new_road_cost_by_type):
        """
        Calculates total road cost, with contribution from new and existing roads. Refer to eq. 3.3.10 in Technical Report.

        Returns
        -------
        float new_and_existing_total_road_cost_USD_df
            Sum total cost in USD of new and existing roads.
        """

        f_new = self.fraction_new_roads    #fraction of new roads that will be built (default assumes 33% new). TODO: Change to user input.
        r_q = self.road_quality    #is a non-dimensional representation of road quality. Default assumes 0.6, which is representative of average road conditions. TODO: Change to user input.
        cost_existing_roads = 0.5 * new_road_cost_by_type      #cost of installing existing roads for the entire project. We currently assume default case to equal 50 % of cost of building new roads.
        new_and_existing_total_road_cost = (f_new * new_road_cost_by_type) + ((1-f_new) * r_q * cost_existing_roads)
        return new_and_existing_total_road_cost


    def calculate_costs(self, calculate_cost_input_dict, calculate_cost_output_dict):
        """
        Function to calculate total labor, equipment, material, mobilization, and anyother associated costs after
        factoring in weather delays.


        Parameters
        ----------
        pd.Dataframe RSMeans
            Dataframe containing labor and equipment rental costs

        pd.DataFrame operation_data
            DataFrame containing estimates for total roads construction time and cost


        Returns
        ----------

        pd.DataFrame total_road_cost
            Dataframe containing following calculated outputs (after weather delay considerations):

            - Total labor cost

            - Totoal material cost

            - Total equipment rental cost

            - Total mobilization cost

            - Any other related costs


        """
        rsmeans = calculate_cost_input_dict['rsmeans']

        material_name = rsmeans['Material type ID'].where(rsmeans['Module'] == 'Roads').dropna().unique()

        material_vol = pd.DataFrame(
            [[material_name[0], calculate_cost_output_dict['material_volume_cubic_yards'], 'Loose cubic yard']],
            columns=['Material type ID', 'Quantity of material', 'Units'])

        material_data = pd.merge(material_vol, calculate_cost_input_dict['material_price'], on=['Material type ID'], sort=True)
        material_data['Cost USD'] = material_data['Quantity of material'] * pd.to_numeric(
            material_data['Material price USD per unit'])

        #New roads material cost:
        material_cost_of_new_roads = material_data['Quantity of material'].iloc[0] * pd.to_numeric(material_data['Material price USD per unit'].iloc[0])

        # New + old roads material cost:
        if calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:
            material_cost_of_old_and_new_roads = self.new_and_existing_total_road_cost(material_cost_of_new_roads)
            material_costs = pd.DataFrame(
                [['Materials', float(material_cost_of_old_and_new_roads), 'Roads']],
                columns=['Type of cost', 'Cost USD', 'Phase of construction'])
        else:
            material_cost_of_old_and_new_roads = material_cost_of_new_roads
            material_costs = pd.DataFrame(
                [['Materials', float(material_cost_of_old_and_new_roads), 'Small DW Roads']],
                columns=['Type of cost', 'Cost USD', 'Phase of construction'])



        operation_data = self.estimate_construction_time(calculate_cost_input_dict, calculate_cost_output_dict)

        wind_delay_fraction = (calculate_cost_output_dict['wind_delay_time'] / calculate_cost_input_dict[
            'operational_hrs_per_day']) / operation_data['Time construct days'].max(skipna=True)
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError('{}: Error: Wind delay greater than 100%'.format(type(self).__name__))
        calculate_cost_output_dict['wind_multiplier'] = 1. / (1. - wind_delay_fraction)

        per_diem = operation_data['Number of workers'] * operation_data['Number of crews'] * (operation_data['Time construct days'] + np.ceil(operation_data['Time construct days'] / 7.)) * calculate_cost_input_dict['rsmeans_per_diem']
        labor_per_diem = per_diem.dropna()

        # calculate_cost_output_dict['Total per diem (USD)'] = per_diem.sum()
        labor_equip_data = pd.merge(operation_data[['Operation ID', 'Units', 'Quantity of material']], rsmeans, on=['Units', 'Operation ID'], sort=True)
        # Calculating labor costs:
        if calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:
            labor_equip_data['Calculated per diem'] = per_diem
            labor_data = labor_equip_data[labor_equip_data['Type of cost'] == 'Labor'].copy()
        else:
            labor_equip_data['Calculated per diem'] = 0.
            # labor_data = labor_equip_data[labor_equip_data['Type of cost'] == 'Labor'].copy()
            labor_data = labor_equip_data[labor_equip_data['Module'] == 'Small DW Roads'].copy()

        overtime_multiplier = calculate_cost_input_dict['overtime_multiplier']
        wind_multiplier = calculate_cost_output_dict['wind_multiplier']

        labor_data['Cost USD'] = ( ((labor_data['Quantity of material'] * labor_data['Rate USD per unit'] *
                                     overtime_multiplier) + labor_per_diem) *
                                   wind_multiplier)

        if calculate_cost_input_dict['road_distributed_wind'] and \
                calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:

            labor_for_new_roads_cost_usd = labor_data['Cost USD'].sum() + \
                                           calculate_cost_output_dict['managament_crew_cost_before_wind_delay']

            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)
            labor_costs = pd.DataFrame([['Labor', float(labor_for_new_and_old_roads_cost_usd), 'Roads']],
                                       columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        elif calculate_cost_input_dict['road_distributed_wind'] and \
                calculate_cost_input_dict['turbine_rating_MW'] < 0.1:  # small DW

            labor_for_new_roads_cost_usd = labor_data['Cost USD'].sum()
            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)

            labor_costs = pd.DataFrame([['Labor', float(labor_for_new_and_old_roads_cost_usd), 'Small DW Roads']],
                                       columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        else:
            labor_for_new_roads_cost_usd = labor_data['Cost USD'].sum() + \
                                           calculate_cost_output_dict['managament_crew_cost_before_wind_delay']

            labor_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(labor_for_new_roads_cost_usd)

            labor_costs = pd.DataFrame([['Labor',
                                         float(labor_for_new_and_old_roads_cost_usd),
                                         'Roads']],

                                       columns=['Type of cost',
                                                'Cost USD',
                                                'Phase of construction'])

        # Filter out equipment costs from rsmeans tab:
        if calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:
            equipment_data = labor_equip_data[labor_equip_data['Type of cost'] == 'Equipment rental'].copy()
        else:
            equipment_data = labor_equip_data[labor_equip_data['Module'] == 'Small DW Roads'].copy()
            equipment_data = equipment_data[equipment_data['Type of cost'] == 'Equipment rental'].copy()

        equipment_data['Cost USD'] = (equipment_data['Quantity of material'] * equipment_data['Rate USD per unit']) * \
                                     calculate_cost_output_dict['wind_multiplier']

        # if rental cost is < half day minimum:
        if equipment_data['Cost USD'].sum() < 460:
            equip_for_new_roads_cost_usd = 460
        else:
            equip_for_new_roads_cost_usd = equipment_data['Cost USD'].sum()

        if calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:
            equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
            equipment_costs = pd.DataFrame([['Equipment rental', float(equip_for_new_and_old_roads_cost_usd), 'Roads']],
                columns=['Type of cost', 'Cost USD', 'Phase of construction'])
        else:
            equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
            equipment_costs = pd.DataFrame([['Equipment rental', float(equip_for_new_and_old_roads_cost_usd), 'Small DW Roads']],
                                           columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # add costs for other operations not included in process data for utility mode (e.g., fencing, access roads)
        if calculate_cost_input_dict['turbine_rating_MW'] > 0.1:
            num_turbines = calculate_cost_input_dict['num_turbines']
            rotor_diameter_m = calculate_cost_input_dict['rotor_diameter_m']
            construct_duration = calculate_cost_input_dict['construct_duration']
            num_access_roads = calculate_cost_input_dict['num_access_roads']

            cost_new_roads_adder = (
                        (float(num_turbines) * 17639) + (float(num_turbines) * float(rotor_diameter_m) * 24.8) + (
                            float(construct_duration) * 55500) + float(num_access_roads) * 3800)
            cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
            additional_costs = pd.DataFrame([['Other', float(cost_adder), 'Roads']],
                                            columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        else:  # No 'Other' cost in distributed wind mode:
            cost_new_roads_adder = 0
            cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
            additional_costs = pd.DataFrame([['Other', float(cost_adder), 'Small DW Roads']],
                                            columns=['Type of cost', 'Cost USD', 'Phase of construction'])



        # Create empty road cost (showing cost breakdown by type) dataframe:
        road_cost = pd.DataFrame(columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        #Filter out equipment costs from rsmeans tab:
        equipment_data = labor_equip_data[labor_equip_data['Type of cost'] == 'Equipment rental'].copy()
        equipment_data['Cost USD'] = (equipment_data['Quantity of material'] * equipment_data['Rate USD per unit']) * calculate_cost_output_dict['wind_multiplier']     #TODO: Annika can you confirm if this is correct.

        equip_for_new_roads_cost_usd = equipment_data['Cost USD'].sum()
        equip_for_new_and_old_roads_cost_usd = self.new_and_existing_total_road_cost(equip_for_new_roads_cost_usd)
        equipment_costs = pd.DataFrame([['Equipment rental', float(equip_for_new_and_old_roads_cost_usd), 'Roads']],
            columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # add costs for other operations not included in process data (e.g., fencing, access roads)
        #
        # Assume the following things to compute the cost adder:
        #
        # The crews only work 6 days per week, so a number of days equal to the
        # number of weeks worked is added to convert working days to calendar
        # days.
        #
        # All crews work concurrently, so the total site preparation time is the
        # the max of "Time construct days"

        max_time_construct_days = operation_data['Time construct days'].max()
        num_turbines = float(calculate_cost_input_dict['num_turbines'])
        rotor_diameter_m = float(calculate_cost_input_dict['rotor_diameter_m'])
        num_access_roads = float(calculate_cost_input_dict['num_access_roads'])
        calendar_construct_days = (max_time_construct_days + np.ceil(
            max_time_construct_days / 6))  # assumes working only 6 days per week
        siteprep_construction_months = calendar_construct_days / 30.0
        cost_new_roads_adder = \
            num_turbines * 17639 + num_turbines * rotor_diameter_m * 24.8 + calculate_cost_input_dict['construct_duration'] * 55500 \
            + num_access_roads * 3800
        cost_adder = self.new_and_existing_total_road_cost(cost_new_roads_adder)
        additional_costs = pd.DataFrame([['Other', cost_adder, 'Roads']],
                                        columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        road_cost = pd.concat((road_cost,material_costs,equipment_costs,labor_costs,additional_costs))

        # set mobilization cost equal to 5% of total road cost for utility scale model and function of
        # of turbine size for distributed wind:
        if calculate_cost_input_dict['num_turbines'] > 10:
            mobilization_costs_new_roads = road_cost["Cost USD"].sum() * 0.05
            mobilization_costs_new_plus_old_roads = self.new_and_existing_total_road_cost(mobilization_costs_new_roads)
            mobilization_costs = pd.DataFrame([['Mobilization', mobilization_costs_new_plus_old_roads, 'Roads']],
                                              columns=['Type of cost', 'Cost USD', 'Phase of construction'])
        else:
            mobilization_costs_new_roads = road_cost["Cost USD"].sum() * \
                                           self.mobilization_cost_multiplier(calculate_cost_input_dict['turbine_rating_MW'])
            mobilization_costs_new_plus_old_roads = self.new_and_existing_total_road_cost(mobilization_costs_new_roads)

            if calculate_cost_input_dict['turbine_rating_MW'] >= 0.1:
                mobilization_costs = pd.DataFrame([['Mobilization', mobilization_costs_new_plus_old_roads, 'Roads']],
                                                  columns=['Type of cost', 'Cost USD', 'Phase of construction'])
            else:
                mobilization_costs = pd.DataFrame([['Mobilization', mobilization_costs_new_plus_old_roads, 'Small DW Roads']],
                                                  columns=['Type of cost', 'Cost USD', 'Phase of construction'])


        road_cost = pd.concat((road_cost, mobilization_costs))
        total_road_cost = road_cost
        calculate_cost_output_dict['total_road_cost'] = total_road_cost
        calculate_cost_output_dict['siteprep_construction_months'] = siteprep_construction_months
        return total_road_cost


    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.


        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        module = type(self).__name__

        # Note that some values are cast with float() so that XlsxWriter
        # (the library depended on by XlsxGenerator) can output them as
        # numbers. XlsxWriter, interestingly, cannot handle np.float32()
        # types.

        result.append({
            'unit': 'm^3',
            'type': 'variable',
            'variable_df_key_col_name': 'Total road volume',
            'value': float(self.output_dict['road_volume_m3'])
        })

        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Depth to subgrade',
            'value': self.output_dict['depth_to_subgrade_m']
        })

        result.append({
            'unit': 'ft',
            'type': 'variable',
            'variable_df_key_col_name': 'Crane path width',
            'value': self.output_dict['crane_path_width_m']     #TODO: Rename variable to: crane_path_width_ft
        })

        if not input_dict['road_distributed_wind']:
            result.append({
                'unit': 'm',
                'type': 'variable',
                'variable_df_key_col_name': 'Road length',
                'value': float(self.output_dict['road_length_m'])
            })

        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Road width',
            'value': self.output_dict['road_width_m']
        })

        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Road thickness',
            'value': self.output_dict['road_thickness_m']
        })



        result.append({
            'unit': 'cubic yards',
            'type': 'variable',
            'variable_df_key_col_name': 'Material volume',
            'value': float(self.output_dict['material_volume_cubic_yards'])
        })

        result.append({
            'unit': 'cubic yards',
            'type': 'variable',
            'variable_df_key_col_name': 'Topsoil volume',
            'value': float(self.output_dict['topsoil_volume'])
        })

        if input_dict['turbine_rating_MW'] >= 0.1:
            result.append({
                'unit': 'cubic yards',
                'type': 'variable',
                'variable_df_key_col_name': 'Embankment volume crane',
                'value': float(self.output_dict['embankment_volume_crane'])
            })

            result.append({
                'unit': 'cubic yards',
                'type': 'variable',
                'variable_df_key_col_name': 'Embankment volume road',
                'value': float(self.output_dict['embankment_volume_road'])
            })

            result.append({
                'unit': 'ft^2',
                'type': 'variable',
                'variable_df_key_col_name': 'Rough grading area',
                'value': float(self.output_dict['rough_grading_area'])
            })

        for row in self.output_dict['total_road_cost'].itertuples():
            dashed_row = '{} <--> {} <--> {}'.format(row[1], row[3], math.ceil(row[2]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost <--> Phase of Construction <--> Cost in USD ',
                'value': dashed_row,
                'last_number': row[2]
            })



        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['roads_cost_csv'] = result
        return result


    def outputs_for_module_type_operation(self, input_dict, output_dict):
        """
        Outputs dictionaries that are rows for the
        costs_by_module_type_operation

        Returns
        -------
        list
            List of dicts, with each dict representing a row for
            the output.
        """
        result = []
        module = type(self).__name__
        num_turbines = self.input_dict['num_turbines']

        costs_by_module_type_operation = output_dict['total_road_cost']
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict['operation_id'] = row['Phase of construction']
            _dict['type_of_cost'] = row['Type of cost']
            _dict['cost'] = row['Cost USD']
            result.append(_dict)

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module
            _dict['total_or_turbine'] = 'total'
            _dict['num_turbines'] = num_turbines

        output_dict['roads_cost_module_type_operation'] = result
        return result

    def run_module(self):
        """
        Runs the SitePreparation module and populates the IO dictionaries with calculated values.

        """
        try:
            self.calculate_road_properties(self.input_dict, self.output_dict)
            operation_data = self.estimate_construction_time(self.input_dict, self.output_dict)

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ('wind_shear_exponent',
                                 'weather_window')

            # specify roads-specific weather delay inputs
            self.weather_input_dict = dict([(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)])
            self.weather_input_dict['start_delay_hours'] = 0  # assume zero start for when road construction begins (start at beginning of construction time)
            self.weather_input_dict[
                'critical_wind_speed_m_per_s'] = self.input_dict['critical_speed_non_erection_wind_delays_m_per_s']
            self.weather_input_dict[
                'wind_height_of_interest_m'] = self.input_dict['critical_height_non_erection_wind_delays_m']

            # compute and specify weather delay mission time for roads
            duration_construction = operation_data['Time construct days'].max(skipna=True)
            operational_hrs_per_day = self.input_dict['hour_day'][self.input_dict['time_construct']]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict['mission_time_hours'] = mission_time_hrs

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict['siteprep_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['total_road_cost'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0  # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} SitePreparationCost")
            return 1, error  # module did not run successfully



import unittest

import numpy as np
import numpy.testing as npt

import wisdem.floatingse.mooring as mm


class TestMooring(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}

        myones = np.ones(1)
        self.inputs["fairlead"] = 10.0 * myones
        self.inputs["fairlead_radius"] = 11.0 * myones
        self.inputs["anchor_radius"] = 250.0 * myones
        self.inputs["anchor_mass"] = 100.0 * myones
        self.inputs["anchor_cost"] = 10.0 * myones
        self.inputs["anchor_max_vertical_load"] = np.array([ 1e10 ])
        self.inputs["anchor_max_lateral_load"] = np.array([ 1e10 ])

        self.inputs["rho_water"] = np.array([ 1e3 ])
        self.inputs["water_depth"] = np.array([ 200.0 ])  # 100.0

        self.inputs["line_length"] = 0.6 * (self.inputs["water_depth"] + self.inputs["anchor_radius"]) * myones
        self.inputs["line_diameter"] = 0.1 * myones
        self.inputs["line_mass_density_coeff"] = 10.0 * myones
        self.inputs["line_stiffness_coeff"] = 2e6 * myones
        self.inputs["line_breaking_load_coeff"] = 30.0 * myones
        self.inputs["line_cost_rate_coeff"] = 40.0 * myones
        self.inputs["max_surge_fraction"] = np.array([ 0.2 ])
        self.inputs["survival_heel"] = np.array([ np.deg2rad(10.0) ])
        self.inputs["operational_heel"] = np.array([ np.deg2rad(5.0) ])

        opt = {}
        opt["n_attach"] = 3
        opt["n_anchors"] = 6
        opt["line_anchor"] = ["custom"]  # ["drag_embedment"]
        opt["line_material"] = ["custom"]  # ["chain"]

        self.mymap = mm.Mooring(options=opt, gamma=1.35)

    def testGeometry(self):
        self.mymap.compute(self.inputs, self.outputs)
        npt.assert_equal(self.outputs["constr_mooring_length"], 0.6 * 450 / (0.95 * 440))

        self.inputs["line_length"] = np.array([ 150.0 ])
        self.mymap.compute(self.inputs, self.outputs)
        npt.assert_equal(self.outputs["constr_mooring_length"], 150 / (200 - 10 - 1.35 * 11 * np.sin(np.deg2rad(10))))

    def testRunMap(self):
        self.mymap.compute(self.inputs, self.outputs)

        npt.assert_almost_equal(self.outputs["mooring_neutral_load"].sum(axis=0)[:2], 0.0, decimal=2)
        npt.assert_equal(self.outputs["mooring_neutral_load"].shape[0], 3)
        npt.assert_array_less(1.0, self.outputs["constr_axial_load"])
        npt.assert_array_less(1e3, self.outputs["max_surge_restoring_force"])

        npt.assert_almost_equal(self.outputs["operational_heel_restoring_force"][[1, 3, 5]], 0.0, 2)
        npt.assert_array_less(0.0, self.outputs["operational_heel_restoring_force"][0])
        npt.assert_array_less(1e3, np.abs(self.outputs["operational_heel_restoring_force"][2]))
        npt.assert_array_less(1e3, np.abs(self.outputs["operational_heel_restoring_force"][4]))

        npt.assert_almost_equal(self.outputs["survival_heel_restoring_force"][[1, 3, 5]], 0.0, 2)
        npt.assert_array_less(
            self.outputs["operational_heel_restoring_force"][0], self.outputs["survival_heel_restoring_force"][0]
        )
        npt.assert_array_less(1e3, np.abs(self.outputs["survival_heel_restoring_force"][2]))
        npt.assert_array_less(
            np.abs(self.outputs["operational_heel_restoring_force"][4]),
            np.abs(self.outputs["survival_heel_restoring_force"][4]),
        )

        npt.assert_almost_equal(self.outputs["mooring_mass"], 6 * 270 * 0.1 + 6 * 100)
        npt.assert_almost_equal(self.outputs["mooring_cost"], 6 * 270 * 0.4 + 6 * 10)

    def testRunMap_MoorProps(self):
        opt = {}
        opt["n_attach"] = 3
        opt["n_anchors"] = 6
        opt["line_anchor"] = ["drag_embedment"]
        opt["line_material"] = ["chain"]

        mymap = mm.Mooring(options=opt, gamma=1.35)

        mymap.compute(self.inputs, self.outputs)
        npt.assert_almost_equal(self.outputs["mooring_mass"], 6 * 270 * 199.0 + 6 * 0)
        npt.assert_almost_equal(self.outputs["mooring_cost"], 6 * 270 * 514.415)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import numpy as np
import numpy.testing as npt

import wisdem.precomp.properties as prop
#from wisdem.precomp._precomp import precomp as _precomp

try:
   import cPickle as pickle
except Exception:
   import pickle

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
# https://stackoverflow.com/questions/20716812/saving-and-loading-multiple-objects-in-pickle-file
def loadall(filename):
    with open(os.path.join(mydir, filename), "rb") as f:
        while True:
            try:
                yield pickle.load(f)
            except EOFError:
                break

class TestPreCompProperties(unittest.TestCase):
    def test_tw_rate(self):

        # From NREL 5 MW
        myr = np.array([ 1.5,         3.62068966,  5.74137931,  7.86206897,  9.98275862, 12.10344828,
                         14.22413793, 16.34482759, 18.46551724, 20.5862069 , 22.70689655, 24.82758621,
                         26.94827586, 29.06896552, 31.18965517, 33.31034483, 35.43103448, 37.55172414,
                         39.67241379, 41.79310345, 43.9137931 , 46.03448276, 48.15517241, 50.27586207,
                         52.39655172, 54.51724138, 56.63793103, 58.75862069, 60.87931034, 63.        ])
        th0 = np.zeros(myr.shape)
        #th_prime_fort = _precomp.tw_rate(myr, th0)
        th_prime_fort = th0
        th_prime_py = prop.tw_rate(myr, th0)
        npt.assert_almost_equal(th_prime_fort, th_prime_py)

        th0 = np.linspace(10., -10., myr.size)
        #th_prime_fort = _precomp.tw_rate(myr, th0)
        th_prime_fort = -0.32520325*np.ones(th0.shape)
        th_prime_py = prop.tw_rate(myr, th0)
        npt.assert_almost_equal(th_prime_fort, th_prime_py)

    def test_properties(self):
        fnames = ['section_dump_nrel5mw.pkl', 'section_dump_iea15mw.pkl']
        
        for f in fnames:
            with self.subTest(f=f):
                myitems = loadall(f)
                nsec = myitems.__next__()
                for k in range(nsec):
                    with self.subTest(i=k):
                        chord = myitems.__next__()
                        theta = myitems.__next__()
                        th_prime = myitems.__next__()
                        le_loc = myitems.__next__()
                        xnode = myitems.__next__()
                        ynode = myitems.__next__()
                        E1 = myitems.__next__()
                        E2 = myitems.__next__()
                        G12 = myitems.__next__()
                        nu12 = myitems.__next__()
                        rho = myitems.__next__()
                        locU = myitems.__next__()
                        n_laminaU = myitems.__next__()
                        n_pliesU = myitems.__next__()
                        tU = myitems.__next__()
                        thetaU = myitems.__next__()
                        mat_idxU = myitems.__next__()
                        locL = myitems.__next__()
                        n_laminaL = myitems.__next__()
                        n_pliesL = myitems.__next__()
                        tL = myitems.__next__()
                        thetaL = myitems.__next__()
                        mat_idxL = myitems.__next__()
                        nwebs = myitems.__next__()
                        locW = myitems.__next__()
                        n_laminaW = myitems.__next__()
                        n_pliesW = myitems.__next__()
                        tW = myitems.__next__()
                        thetaW = myitems.__next__()
                        mat_idxW = myitems.__next__()

                        results_fort = myitems.__next__()
                        results_py = prop.properties(chord,
                                                     theta,
                                                     th_prime,
                                                     le_loc,
                                                     xnode,
                                                     ynode,
                                                     E1,
                                                     E2,
                                                     G12,
                                                     nu12,
                                                     rho,
                                                     locU,
                                                     n_laminaU,
                                                     n_pliesU,
                                                     tU,
                                                     thetaU,
                                                     mat_idxU,
                                                     locL,
                                                     n_laminaL,
                                                     n_pliesL,
                                                     tL,
                                                     thetaL,
                                                     mat_idxL,
                                                     nwebs,
                                                     locW,
                                                     n_laminaW,
                                                     n_pliesW,
                                                     tW,
                                                     thetaW,
                                                     mat_idxW,
                                                     )

                        npt.assert_almost_equal(results_fort, results_py, decimal=5)


if __name__ == "__main__":
    unittest.main()



"""Provides the `MonopileDesign` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from math import pi, log

from scipy.optimize import fsolve

from wisdem.orbit.core.defaults import common_costs
from wisdem.orbit.phases.design import DesignPhase


class MonopileDesign(DesignPhase):
    """Monopile Design Class."""

    expected_config = {
        "site": {"depth": "m", "mean_windspeed": "m/s"},
        "plant": {"num_turbines": "int"},
        "turbine": {
            "rotor_diameter": "m",
            "hub_height": "m",
            "rated_windspeed": "m/s",
        },
        "monopile_design": {
            "yield_stress": "Pa (optional)",
            "load_factor": "float (optional)",
            "material_factor": "float (optional)",
            "monopile_density": "kg/m3 (optional)",
            "monopile_modulus": "Pa (optional)",
            "monopile_tp_connection_thickness": "m (optional)",
            "transition_piece_density": "kg/m3 (optional)",
            "transition_piece_thickness": "m (optional)",
            "transition_piece_length": "m (optional)",
            "soil_coefficient": "N/m3 (optional)",
            "air_density": "kg/m3 (optional)",
            "weibull_scale_factor": "float (optional)",
            "weibull_shape_factor": "float (optional)",
            "turb_length_scale": "m (optional)",
            "monopile_steel_cost": "USD/t (optional)",
            "tp_steel_cost": "USD/t (optional)",
        },
    }

    output_config = {
        "monopile": {
            "diameter": "m",
            "thickness": "m",
            "moment": "m4",
            "embedment_length": "m",
            "length": "m",
            "mass": "t",
            "deck_space": "m2",
            "unit_cost": "USD",
        },
        "transition_piece": {
            "length": "m",
            "mass": "t",
            "deck_space": "m2",
            "unit_cost": "USD",
        },
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of MonopileDesign.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self._outputs = {}

    def run(self):
        """
        Main run function. Passes required config parameters to
        :py:meth:`.design_monopile`.
        """

        _kwargs = self.config.get("monopile_design", {})
        self._outputs["monopile"] = self.design_monopile(
            mean_windspeed=self.config["site"]["mean_windspeed"],
            site_depth=self.config["site"]["depth"],
            rotor_diameter=self.config["turbine"]["rotor_diameter"],
            hub_height=self.config["turbine"]["hub_height"],
            rated_windspeed=self.config["turbine"]["rated_windspeed"],
            **_kwargs,
        )

        self._outputs["transition_piece"] = self.design_transition_piece(
            self.monopile_sizing["diameter"],
            self.monopile_sizing["thickness"],
            **_kwargs,
        )

    def design_monopile(
        self,
        mean_windspeed,
        site_depth,
        rotor_diameter,
        hub_height,
        rated_windspeed,
        **kwargs,
    ):
        """
        Solves system of equations for the required pile diameter to satisfy
        the 50 year extreme operating gust moment. Using the result from the
        diameter equation, calculates the wall thickness and the required
        embedment length and other important sizing parameters.

        Parameters
        ----------
        mean_windspeed : int | float
            Mean wind speed at site (m/s).
        site_depth : int | float
            Water depth at site (m).
        rotor_diameter : int | float
            Rotor diameter (m).
        hub_height : int | float
            Hub height above mean sea level (m).
        rated_windspeed : int | float
            Rated windspeed of turbine (m/s).

        Returns
        -------
        monopile : dict
            Dictionary of monopile sizing and costs.

            - ``diameter`` - Pile diameter (m)
            - ``thickness`` - Pile wall thickness (m)
            - ``moment`` - Pile bending moment of inertia (m4)
            - ``embedment_length`` - Pile embedment length (m)
            - ``length`` - Total pile length (m)
            - ``mass`` - Pile mass (t)
            - ``type`` - `'Monopile'`

        References
        ----------
        This class was adapted from [#arany2017]_.

        .. [#arany2017] Laszlo Arany, S. Bhattacharya, John Macdonald,
           S.J. Hogan, Design of monopiles for offshore wind turbines in 10
           steps, Soil Dynamics and Earthquake Engineering,
           Volume 92, 2017, Pages 126-152, ISSN 0267-7261,
        """

        yield_stress = kwargs.get("yield_stress", 355000000)  # PA
        material_factor = kwargs.get("material_factor", 1.1)
        M_50y = self.calculate_50year_wind_moment(
            mean_windspeed=mean_windspeed,
            site_depth=site_depth,
            rotor_diameter=rotor_diameter,
            hub_height=hub_height,
            rated_windspeed=rated_windspeed,
            **kwargs,
        )

        data = (yield_stress, material_factor, M_50y)
        monopile = {}

        # Monopile sizing
        monopile["diameter"] = fsolve(self.pile_diam_equation, 10, args=data)[0]
        monopile["thickness"] = self.pile_thickness(monopile["diameter"])
        monopile["moment"] = self.pile_moment(monopile["diameter"], monopile["thickness"])
        monopile["embedment_length"] = self.pile_embedment_length(monopile["moment"], **kwargs)

        # Total length
        airgap = kwargs.get("airgap", 10)  # m
        monopile["length"] = monopile["embedment_length"] + site_depth + airgap
        monopile["mass"] = self.pile_mass(
            Dp=monopile["diameter"],
            tp=monopile["thickness"],
            Lt=monopile["length"],
            **kwargs,
        )

        # Deck space
        monopile["deck_space"] = monopile["diameter"] ** 2

        # Costs
        monopile["unit_cost"] = monopile["mass"] * self.monopile_steel_cost

        self.monopile_sizing = monopile
        return monopile

    def design_transition_piece(self, D_p, t_p, **kwargs):
        """
        Designs transition piece given the results of the monopile design.

        Based on Arany 2016, sections 2.2.7 - 2.2.8.

        Parameters
        ----------
        monopile_diameter : int | float
            Diameter of the designed monopile.

        Returns
        -------
        tp_design : dict
            Transition piece design parameters.
        """

        # Defaults to a bolted connection
        t_c = kwargs.get("monopile_tp_connection_thickness", 0.0)

        dens_tp = kwargs.get("transition_piece_density", 7860)  # kg/m3
        t_tp = kwargs.get("transition_piece_thickness", t_p)
        L_tp = kwargs.get("transition_piece_length", 25)
        D_tp = D_p + 2 * (t_c + t_tp)  # Arany 2016, Section 2.2.7

        # Arany 2016, Section 2.2.8
        m_tp = (dens_tp * (D_p + 2 * t_c + t_tp) * pi * t_tp * L_tp) / 907.185  # t

        tp_design = {
            "thickness": t_tp,
            "diameter": D_tp,
            "mass": m_tp,
            "length": L_tp,
            "deck_space": D_tp**2,
            "unit_cost": m_tp * self.tp_steel_cost,
        }

        return tp_design

    @property
    def design_result(self):
        """Returns the results of :py:meth:`.design_monopile`."""

        if not self._outputs:
            raise Exception("Has MonopileDesign been ran yet?")

        return self._outputs

    @property
    def total_cost(self):
        """Returns total cost of the substructures and transition pieces."""

        return sum([v for _, v in self.material_cost.items()])

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        _outputs = {
            "total_monopile_mass": self.total_monopile_mass,
            "total_monopile_cost": self.material_cost["monopile"],
            "total_transition_piece_mass": self.total_tp_mass,
            "total_transition_piece_cost": self.material_cost["transition_piece"],
        }

        return _outputs

    @property
    def material_cost(self):
        """Returns the material cost of the monopile and transition piece."""

        if not self._outputs:
            raise Exception("Has MonopileDesign been ran yet?")

        out = {
            "monopile": self.total_monopile_mass * self.monopile_steel_cost,
            "transition_piece": self.total_tp_mass * self.tp_steel_cost,
        }

        return out

    @property
    def total_monopile_mass(self):
        """Returns total mass of all monopiles."""

        if not self._outputs:
            raise Exception("Has MonopileDesign been ran yet?")

        num_turbines = self.config["plant"]["num_turbines"]

        return self._outputs["monopile"]["mass"] * num_turbines

    @property
    def total_tp_mass(self):
        """Returns total mass of all transition pieces."""

        if not self._outputs:
            raise Exception("Has MonopileDesign been ran yet?")

        num_turbines = self.config["plant"]["num_turbines"]

        return self._outputs["transition_piece"]["mass"] * num_turbines

    @property
    def monopile_steel_cost(self):
        """Returns the cost of monopile steel (USD/t) fully fabricated."""

        _design = self.config.get("monopile_design", {})
        _key = "monopile_steel_cost"

        try:
            cost = _design.get(_key, common_costs[_key])

        except KeyError:
            raise Exception("Cost of monopile steel not found.")

        return cost

    @property
    def tp_steel_cost(self):
        """
        Returns the cost of transition piece steel (USD/t) fully fabricated.
        """

        _design = self.config.get("monopile_design", {})
        _key = "tp_steel_cost"

        try:
            cost = _design.get(_key, common_costs[_key])

        except KeyError:
            raise Exception("Cost of transition piece steel not found.")

        return cost

    @staticmethod
    def pile_mass(Dp, tp, Lt, **kwargs):
        """
        Calculates the total monopile mass in tonnes.

        Parameters
        ----------
        Dp : int | float
            Pile diameter (m).
        tp : int | float
            Pile wall thickness (m).
        Lt : int | float
            Total pile length (m).

        Returns
        -------
        mt : float
            Total pile mass (t).
        """

        density = kwargs.get("monopile_density", 7860)  # kg/m3
        volume = (pi / 4) * (Dp**2 - (Dp - tp) ** 2) * Lt
        mass = density * volume / 907.185

        return mass

    @staticmethod
    def pile_embedment_length(Ip, **kwargs):
        """
        Calculates required pile embedment length.
        Source: Arany & Bhattacharya (2016)
        - Equation 7 (Enforces a rigid/lower aspect ratio monopile)

        Parameters
        ----------
        Ip : int | float
            Pile moment of inertia (m4)

        Returns
        -------
        Lp : float
            Required pile embedment length (m).
        """

        monopile_modulus = kwargs.get("monopile_modulus", 200e9)  # Pa
        soil_coefficient = kwargs.get("soil_coefficient", 4000000)  # N/m3

        Lp = 2 * ((monopile_modulus * Ip) / soil_coefficient) ** 0.2

        return Lp

    @staticmethod
    def pile_thickness(Dp):
        """
        Calculates pile wall thickness.
        Source: Arany & Bhattacharya (2016)
        - Equation 1

        Parameters
        ----------
        Dp : int | float
            Pile diameter (m).

        Returns
        -------
        tp : float
            Pile Wall Thickness (m)
        """

        tp = 0.00635 + Dp / 100

        return tp

    @staticmethod
    def pile_moment(Dp, tp):
        """
        Equation to calculate the pile bending moment of inertia.

        Parameters
        ----------
        Dp : int | float
            Pile diameter (m).
        tp : int | float
            Pile wall thickness (m).

        Returns
        -------
        Ip : float
            Pile bending moment of inertia
        """

        Ip = 0.125 * ((Dp - tp) ** 3) * tp * pi

        return Ip

    @staticmethod
    def pile_diam_equation(Dp, *data):
        """
        Equation to be solved for Pile Diameter. Combination of equations 99 &
        101 in this paper:
        Source: Arany & Bhattacharya (2016)
        - Equations 99 & 101

        Parameters
        ----------
        Dp : int | float
            Pile diameter (m).

        Returns
        -------
        res : float
            Reduced equation result.
        """

        yield_stress, material_factor, M_50y = data
        A = (yield_stress * pi) / (4 * material_factor * M_50y)
        res = A * ((0.99 * Dp - 0.00635) ** 3) * (0.00635 + 0.01 * Dp) - Dp

        return res

    def calculate_50year_wind_moment(
        self,
        mean_windspeed,
        site_depth,
        rotor_diameter,
        hub_height,
        rated_windspeed,
        **kwargs,
    ):
        """
        Calculates the 50 year extreme wind moment using methodology from
        DNV-GL. Source: Arany & Bhattacharya (2016)
        - Equation 30

        Parameters
        ----------
        mean_windspeed : int | float
            Mean wind speed at site (m/s).
        site_depth : int | float
            Water depth at site (m).
        rotor_diameter : int | float
            Rotor diameter (m).
        hub_height : int | float
            Hub height above mean sea level (m).
        rated_windspeed : int | float
            Rated windspeed of turbine (m/s).
        load_factor : float
            Added safety factor on the extreme wind moment.
            Default: 3.375 (2.5x DNV standard as this model does not design for buckling or fatigue)

        Returns
        -------
        M_50y : float
            50 year extreme wind moment (N-m).
        """

        load_factor = kwargs.get("load_factor", 3.375)

        F_50y = self.calculate_50year_wind_load(
            mean_windspeed=mean_windspeed,
            rotor_diameter=rotor_diameter,
            rated_windspeed=rated_windspeed,
            **kwargs,
        )

        M_50y = F_50y * (site_depth + hub_height)

        return M_50y * load_factor

    def calculate_50year_wind_load(self, mean_windspeed, rotor_diameter, rated_windspeed, **kwargs):
        """
        Calculates the 50 year extreme wind load using methodology from DNV-GL.
        Source: Arany & Bhattacharya (2016)
        - Equation 29

        Parameters
        ----------
        mean_windspeed : int | float
            Mean wind speed at site (m/s).
        rotor_diam : int | float
            Rotor diameter (m).
        rated_windspeed : int | float
            Rated windspeed of turbine (m/s).

        Returns
        -------
        F_50y : float
            50 year extreme wind load (N).
        """

        dens = kwargs.get("air_density", 1.225)
        swept_area = pi * (rotor_diameter / 2) ** 2

        ct = self.calculate_thrust_coefficient(rated_windspeed=rated_windspeed)

        U_eog = self.calculate_50year_extreme_gust(
            mean_windspeed=mean_windspeed,
            rated_windspeed=rated_windspeed,
            rotor_diameter=rotor_diameter,
            **kwargs,
        )

        F_50y = 0.5 * dens * swept_area * ct * (rated_windspeed + U_eog) ** 2

        return F_50y

    @staticmethod
    def calculate_thrust_coefficient(rated_windspeed):
        """
        Calculates the thrust coefficient using rated windspeed.
        Source: Frohboese & Schmuck (2010)

        Parameters
        ----------
        rated_windspeed : int | float
            Rated windspeed of turbine (m/s).

        Returns
        -------
        ct : float
            Coefficient of thrust.
        """

        ct = min([3.5 * (2 * rated_windspeed + 3.5) / (rated_windspeed**2), 1])

        return ct

    @staticmethod
    def calculate_50year_extreme_ws(mean_windspeed, **kwargs):
        """
        Calculates the 50 year extreme wind speed using methodology from DNV-GL.
        Source: Arany & Bhattacharya (2016)
        - Equation 27

        Parameters
        ----------
        mean_windspeed : int | float
            Mean wind speed (m/s).
        shape_factor : int | float
            Shape factor of the Weibull distribution.

        Returns
        -------
        U_50y : float
            50 year extreme wind speed (m/s).
        """

        scale_factor = kwargs.get("weibull_scale_factor", mean_windspeed)
        shape_factor = kwargs.get("weibull_shape_factor", 2)
        U_50y = scale_factor * (-log(1 - 0.98 ** (1 / 52596))) ** (1 / shape_factor)

        return U_50y

    def calculate_50year_extreme_gust(self, mean_windspeed, rotor_diameter, rated_windspeed, **kwargs):
        """
        Calculates the 50 year extreme wind gust using methodology from DNV-GL.
        Source: Arany & Bhattacharya (2016)
        - Equation 28

        Parameters
        ----------
        mean_windspeed : int | float
            Mean wind speed at site (m/s).
        rotor_diameter : int | float
            Rotor diameter (m).
        rated_windspeed : int | float
            Rated windspeed of turbine (m/s).
        turb_length_scale : int | float
            Turbulence integral length scale (m).

        Returns
        -------
        U_eog : float
            Extreme operating gust speed (m/s).
        """

        length_scale = kwargs.get("turb_length_scale", 340.2)

        U_50y = self.calculate_50year_extreme_ws(mean_windspeed, **kwargs)
        U_1y = 0.8 * U_50y

        U_eog = min(
            [
                (1.35 * (U_1y - rated_windspeed)),
                (3.3 * 0.11 * U_1y) / (1 + (0.1 * rotor_diameter) / (length_scale / 8)),
            ]
        )

        return U_eog
